import numpy as np
import utool as ut
import ubelt as ub
import scipy.interpolate
from functools import partial


def check_unused_kwargs(kwargs, expected_keys):
    unused_keys = set(kwargs.keys()) - set(expected_keys)
    if len(unused_keys) > 0:
        print('unused kwargs keys = %r' % (unused_keys))


def testdata_score_normalier(tp_bumps=[(6.5, 256)], tn_bumps=[(3.5, 256)], tp_scale=1.0, tn_scale=1.0, min_clip=None, **kwargs):
    rng = np.random.RandomState(seed=0)
    # Get a training sample
    tp_support = np.hstack([rng.normal(loc=loc, scale=tp_scale, size=(size,)) for loc, size in tp_bumps])
    tn_support = np.hstack([rng.normal(loc=loc, scale=tn_scale, size=(size,)) for loc, size in tn_bumps])
    if min_clip is not None:
        tp_support[tp_support < min_clip] = min_clip
        tn_support[tn_support < min_clip] = min_clip
    data   = np.hstack((tp_support, tn_support))
    labels = np.array([True] * len(tp_support) + [False] * len(tn_support))
    encoder = ScoreNormalizer(**kwargs)
    encoder.fit(data, labels)
    return encoder, data, labels


def get_left_area(ydata, xdata, index_list):
    """ area to the left of each index point """
    left_area = np.array([np.trapz(ydata[:ix + 1], xdata[:ix + 1])
                          for ix in index_list])
    return left_area


def get_right_area(ydata, xdata, index_list):
    """ area to the right of each index point """
    right_area = np.array([np.trapz(ydata[ix:], xdata[ix:])
                           for ix in index_list])
    return right_area


class ScoreNormVisualizeClass(object):
    """
    # HACK; eventually move all individual plots into a class structure
    """

    def _hack_vizlearn(encoder, **kwargs):
        if 'target_tpr' in kwargs:
            print('_HACK VIZLERAN TARGET')
            verbose = ut.VERBOSE
            score_thresh = encoder.learn_threshold(verbose=verbose, **kwargs)
            prob_thresh = encoder.learned_thresh
            #prob_thresh = encoder.normalize_scores(score_thresh)
        else:
            #prob_thresh = encoder.learned_thresh
            #score_thresh = encoder.inverse_normalize(prob_thresh)
            print('_HACK VIZLERAN2')
            score_thresh = encoder.learn_threshold2()
            prob_thresh = encoder.normalize_scores(score_thresh)
        return score_thresh, prob_thresh

    def _plot_score_support_hist(encoder, fnum, pnum=(1, 1, 1), **kwargs):
        import plottool_ibeis as pt
        fnum = pt.ensure_fnum(fnum)
        tup = encoder.get_partitioned_support()
        tp_support, tn_support, part_attrs = tup
        score_thresh, prob_thresh = encoder._hack_vizlearn(**kwargs)

        true_color = pt.TRUE_BLUE  # pt.TRUE_GREEN
        false_color = pt.FALSE_RED
        support_kw = dict(
            score_lbls=('trueneg', 'truepos'),
            score_colors=(false_color, true_color),
            titlesuf=kwargs.get('titlesuf', '')
        )
        score_range = kwargs.get('score_range', None)
        pt.plot_score_histograms(
            (tn_support, tp_support),
            score_thresh=score_thresh,
            score_label='score',
            fnum=fnum,
            pnum=pnum,
            bin_width=kwargs.get('bin_width', None),
            num_bins=kwargs.get('num_bins', None),
            overlay_prob_given_list=(encoder.p_score_given_tn, encoder.p_score_given_tp),
            overlay_score_domain=encoder.score_domain,
            xlim=score_range,
            **support_kw)

    def _plot_roc(encoder, fnum, pnum, **kwargs):
        import vtool_ibeis as vt
        import plottool_ibeis as pt  # NOQA
        tup = encoder.get_partitioned_support()
        tp_support, tn_support, part_attrs = tup

        scores = np.hstack([tn_support, tp_support])
        labels = np.array([False] * len(tn_support) + [True] * len(tp_support))

        # probs = encoder.normalize_scores(scores)
        probs = normalize_scores(encoder.score_domain, encoder.p_tp_given_score, scores)

        confusions = vt.ConfusionMetrics().fit(probs, labels)

        score_thresh, prob_thresh = encoder._hack_vizlearn(**kwargs)

        #target_tpr = None
        target_tpr = confusions.get_metric_at_thresh('tpr', prob_thresh)
        #print('target_tpr = %r' % (target_tpr,))
        ROCInteraction = vt.interact_roc_factory(confusions, target_tpr,
                                                 show_operating_point=True)
        fnum = pt.ensure_fnum(fnum)
        ROCInteraction.static_plot(fnum, pnum, **kwargs)

    def _plot_prebayes(encoder, fnum, pnum, **kwargs):
        score_thresh, prob_thresh = encoder._hack_vizlearn(**kwargs)
        plot_prebayes_pdf(
            encoder.score_domain, encoder.p_score_given_tn, encoder.p_score_given_tp, encoder.p_score,
            score_thresh=score_thresh,
            cfgstr='', fnum=fnum, pnum=pnum)

    def _plot_postbayes(encoder, fnum, pnum, **kwargs):
        score_thresh, prob_thresh = encoder._hack_vizlearn(**kwargs)
        plot_postbayes_pdf(encoder.score_domain, encoder.p_tn_given_score, encoder.p_tp_given_score,
                           prob_thresh=prob_thresh,
                           score_thresh=score_thresh,
                           cfgstr='', fnum=fnum, pnum=pnum)


class ScoreNormalizer(ut.Cachable, ScoreNormVisualizeClass):
    """
    Conforms to scikit-learn Estimator interface

    CommandLine:
        python -m vtool_ibeis.score_normalization --test-ScoreNormalizer --show --cmd

    Kwargs:
        tpr (float): target true positive rate (default .90)
        fpr (float): target false positive rate (default None)
        gridsize=1024,
        adjust=8,
        monotonize=False, if True ensures inferred probability curves are monotonic
        clip_factor=None,
        reverse (bool): True if lower scores are better, False if higher scores
            are better (default=None)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.score_normalization import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> encoder = ScoreNormalizer()
        >>> X, y = vt.demodata.testdata_binary_scores()
        >>> attrs = {'index': np.arange(len(y)) * ((2 * y) - 1)}
        >>> encoder.fit(X, y, attrs)
        >>> # xdoctest: +REQUIRES(--show)
        >>> encoder.visualize()
        >>> ut.show_if_requested()
    """
    def __init__(encoder, **kwargs):
        encoder.learn_kw = ut.update_existing(
            dict(
                gridsize=1024,
                adjust=8,
                monotonize=False,
                #monotonize=True,
                #clip_factor=(ut.PHI + 1),
                clip_factor=None,
                reverse=None,
                p_tp_method='eq',
            ), kwargs)
        #check_unused_kwargs(kwargs, encoder.learn_kw.keys())
        encoder.thresh_kw = ut.update_existing(
            dict(
                # Target recall for learned threshold
                tpr=None,
                fpr=None,
            ), kwargs)
        assert not any(key.startswith('target_') for key in kwargs), (
            'old interface of target_<metric> used just use <metric>')
        if not any(encoder.thresh_kw.values()):
            encoder.thresh_kw['tpr'] = .90
        # Support data
        encoder.support = dict(
            X=None,
            y=None,
            attrs=None,
        )
        # Learned score normalization
        encoder.score_domain     = None
        encoder.p_tp_given_score = None
        encoder.p_tn_given_score = None
        encoder.p_score_given_tn = None
        encoder.p_score_given_tp = None
        encoder.p_score = None
        # Learneed classification threshold
        encoder.learned_thresh   = None
        # Learned interpolation function
        encoder.interp_fn = None

    def __getstate__(encoder):
        """

        CommandLine:
            python -m vtool_ibeis.score_normalization --test-__getstate__

        Example:
            >>> # ENABLE_DOCTEST
            >>> from vtool_ibeis.score_normalization import *  # NOQA
            >>> encoder = ScoreNormalizer()
            >>> from six.moves import cPickle as pickle
            >>> dump = pickle.dumps(encoder)
            >>> encoder2 = pickle.loads(dump)
        """
        state_dict = encoder.__dict__.copy()
        state_dict['interp_fn'] = None
        return state_dict

    def get_prefix(encoder):
        return 'ScoreNorm_'

    def __setstate__(encoder, state_dict):
        encoder.__dict__.update(state_dict)
        encoder._update_interp_fn()

    def fit(encoder, X, y, attrs=None, verbose=False, finite_only=True):
        """
        Fits estimator to data

        Args:
            X (ndarray): one dimensional scores
            y (ndarray): binary labels
            attrs (dict): dictionary of data attributes
        """
        # Record support
        encoder.support['X'] = X
        encoder.support['y'] = y
        encoder.support['attrs'] = {} if attrs is None else attrs
        encoder.learn_probabilities(verbose=verbose)
        try:
            encoder.learn_threshold(verbose=verbose)
        except Exception as ex:
            ut.printex(ex, 'could not learn thresh', iswarning=True)

    @staticmethod
    def _to_xy(tp_scores, tn_scores, part_attrs=None):
        return flatten_scores(tp_scores, tn_scores, part_attrs)

    @staticmethod
    def _to_partitioned(X, y, attrs={}):
        return partition_scores(X, y, attrs)

    def fit_partitioned(encoder, tp_scores, tn_scores, part_attrs=None,
                        **kwargs):
        """ convinience func to fit only scores that have been separated
        instead of labeled"""
        fitargs = flatten_scores(tp_scores, tn_scores, part_attrs)
        return encoder.fit(*fitargs, **kwargs)

    def get_partitioned_support(encoder):
        """ convinience get prepartitioned data """
        X, y, attrs = encoder.get_support()
        return partition_scores(X, y, attrs)

    def get_support(encoder, finite_only=True):
        """
        return X, y, and attrs
        """
        X = encoder.support['X']
        y = encoder.support['y']
        attrs = encoder.support['attrs']
        if finite_only:
            mask = np.isfinite(X)
            X = X.compress(mask)
            y = y.compress(mask)
            attrs = ut.map_dict_vals(partial(np.compress, mask), attrs)
        return X, y, attrs

    def learn_probabilities(encoder, verbose=False):
        """
        Kernel density estimation
        """
        #X, y = encoder.get_support()
        tp_support, tn_support, part_attrs = encoder.get_partitioned_support()
        # heuristic
        encoder.learn_kw['reverse'] = tp_support.mean() < tn_support.mean()
        if verbose:
            print('[scorenorm] setting reverse = %r' %
                  (encoder.learn_kw['reverse']))

        tup = learn_score_normalization(tp_support, tn_support, return_all=True,
                                        verbose=verbose, **encoder.learn_kw)
        # unpack
        (score_domain, p_tp_given_score, p_tn_given_score,
         p_score_given_tp, p_score_given_tn, p_score) = tup

        encoder.score_domain = score_domain
        encoder.p_tp_given_score = p_tp_given_score
        encoder.p_tn_given_score = p_tn_given_score
        encoder.p_score_given_tn = p_score_given_tn
        encoder.p_score_given_tp = p_score_given_tp
        encoder.p_score = p_score
        encoder._update_interp_fn()

    def _update_interp_fn(encoder):
        """
        Internal call to update interpolation function. Used when learning and
        when loading from cache.
        """
        if encoder.p_tp_given_score is not None:
            encoder.interp_fn = scipy.interpolate.interp1d(
                encoder.score_domain, encoder.p_tp_given_score, kind='linear',
                copy=False, assume_sorted=False)

    def learn_threshold2(encoder):
        """
        Finds a cutoff where the probability of a truepos stats becoming
        greater than probability of trueneg

        CommandLine:
            python -m vtool_ibeis.score_normalization --exec-learn_threshold2 --show

        Example:
            >>> from vtool_ibeis.score_normalization import *  # NOQA
            >>> import vtool_ibeis as vt
            >>> #encoder, X, y = testdata_score_normalier([(3.5, 256), (9.5, 1024), (15.5, 2048)], [(6.5, 256), (12.5, 5064), (18.5, 128)], adjust=1, p_tp_method='ratio')
            >>> encoder, X, y = testdata_score_normalier([(3.5, 64), (9.5, 1024), (15.5, 5064)], [(6.5, 256), (12.5, 2048), (18.5, 128)], adjust=1, p_tp_method='ratio')
            >>> #encoder, X, y = testdata_score_normalier(adjust=1)
            >>> #encoder, X, y = testdata_score_normalier([(3.5, 2048)], [(30.5, 128)], tn_scale=.1, adjust=1)
            >>> #encoder, X, y = testdata_score_normalier([(0, 64)], [(-.1, 12)], adjust=8, min_clip=0)
            >>> locals_ = ut.exec_func_src(encoder.learn_threshold2)
            >>> exec(ut.execstr_dict(locals_))
            >>> # xdoctest: +REQUIRES(--show)
            >>> import plottool_ibeis as pt
            >>> pt.ensureqt()
            >>> #pt.plot(xdata[0:-2], np.diff(np.diff(closeness)))
            >>> #maxima_x, maxima_y, argmaxima = vt.hist_argmaxima(closeness)
            >>> fnum = 100
            >>> pt.multi_plot(xdata, [tp_curve, tn_curve, closeness, ],
            >>>               label_list=['p(tp | s)', 'p(tn | s)', 'closeness', ], marker='',
            >>>               linewidth_list=[4, 4, 1,], title='intersection points',
            >>>               pnum=(4, 1, 1), fnum=fnum, xmax=xdata.max(), xmin=0)
            >>> pt.plot(xdata[argmaxima], closeness[argmaxima], 'rx', label='closeness maxima')
            >>> pt.plot(x_submax, y_submax, 'o', label='chosen')
            >>> #pt.plot(xdata[argmaxima], curveness[argmaxima], 'rx', label='curveness maxima')
            >>> pt.legend()
            >>> #pt.plot(x_submax, y_submax, 'o')
            >>> pt.plot(xdata[argmaxima], tp_curve[argmaxima], 'rx')
            >>> pt.plot(xdata[argmaxima], tn_curve[argmaxima], 'rx')
            >>> pt.plot(xdata[argmaxima], tp_curve[argmaxima], 'rx')
            >>> pt.plot(xdata[argmaxima], tn_curve[argmaxima], 'rx')
            >>> #pt.plot(xdata[argmaxima], encoder.interp_fn(x_submax), 'rx')
            >>> _mkinterp = ut.partial(
            >>>     scipy.interpolate.interp1d, kind='linear', copy=False,
            >>>     assume_sorted=False, bounds_error=False)
            >>> _interp_sgtn = _mkinterp(xdata, tn_curve)
            >>> _interp_sgtp = _mkinterp(xdata, tp_curve)
            >>> pt.plot(x_submax, _interp_sgtn(x_submax), 'go')
            >>> pt.plot(x_submax, _interp_sgtp(x_submax), 'bx')
            >>> #
            >>> pt.multi_plot(xdata[argmaxima], [tp_area, fp_area, tn_area, fn_area], title='intersection areas',
            >>>               label_list=['tp_area', 'fp_area', 'tn_area', 'fn_area'], markers=['o', 'd', 'o', '.'],
            >>>               pnum=(4, 1, 2), fnum=fnum, xmax=xdata.max(), xmin=0)
            >>> #
            >>> pt.multi_plot(xdata[argmaxima], [lr_pos, lr_neg, acc], title='intersection quality (liklihood ratios)',
            >>>               label_list=['lr_pos=tp/fp', 'lr_neg=fn/tn', 'acc'], markers=['o', 'o', '*'],
            >>>               pnum=(4, 1, 3), fnum=fnum, xmax=xdata.max(), xmin=0)
            >>> #
            >>> pnum_ = pt.make_pnum_nextgen(4, 3, start=9)
            >>> encoder._plot_score_support_hist(fnum=fnum, pnum=pnum_())
            >>> #encoder._plot_prebayes(fnum=fnum, pnum=pnum_())
            >>> encoder._plot_postbayes(fnum=fnum, pnum=pnum_())
            >>> encoder._plot_roc(fnum=fnum, pnum=pnum_())
            >>> pt.adjust_subplots(hspace=.5, top=.95, bottom=.08)
            >>> pt.show_if_requested()
        """
        import vtool_ibeis as vt
        if False:
            # New stuff with area should make this irrelevant
            #weights = encoder.p_score_given_tp
            weights = encoder.p_score_given_tn
            values = encoder.score_domain
            mean = np.average(values, weights=weights)
            std = np.sqrt(np.average((values - mean) ** 2, weights=weights))
            score_cutoff = mean + (2.5 * std)
            cutx = np.sum(encoder.score_domain < score_cutoff)
            xdata = encoder.score_domain[:cutx]
            tp_curve = encoder.p_score_given_tp[:cutx]
            tn_curve = encoder.p_score_given_tn[:cutx]
        else:
            xdata = encoder.score_domain
            tp_curve = encoder.p_score_given_tp
            tn_curve = encoder.p_score_given_tn
            if 0:
                _p = np.where(xdata > 50)[0][0]
                xdata = xdata[0:_p]
                tp_curve = tp_curve[0:_p]
                tn_curve = tn_curve[0:_p]
            #print('xdata = %r' % (xdata,))
            #print('tp_curve = %r' % (tp_curve,))
            #print('tn_curve = %r' % (tn_curve,))
            #tp_curve[:] = .1
            #tn_curve[:] = .5

        # Find locations of intersection
        closeness = -np.abs(tp_curve - tn_curve)
        closeness = closeness - closeness.min()
        #closeness
        #print('closeness = %r' % (closeness,))
        #argmaxima = vt.hist_argmaxima2(closeness)
        #print('argmaxima = %r' % (argmaxima,))

        argmaxima = vt.hist_argmaxima2(closeness)
        #argmaxima = np.arange(2, len(closeness) - 2)
        #curvature = -np.gradient(np.gradient(closeness))
        #curveness = (curvature - curvature.min()) / (curvature.max() - curvature.min())
        # Remove maxima points with almost no curvature
        #if False:
        #    if len(argmaxima) > 1:
        #        #curveness[argmaxima]
        #        curvature_  = curvature[argmaxima]
        #        valid = curvature_ > 1e-5
        #        #valid = curvature[argmaxima] > np.median(curvature[argmaxima])
        #        #valid = curvature[argmaxima] > np.median(curvature[argmaxima])
        #        if np.any(valid):
        #            argmaxima = argmaxima[valid]

        #argmaxima2 = vt.hist_argmaxima2(-deriv_no2)
        #if len(np.intersect1d(argmaxima2, argmaxima)) > 0:
        #    argmaxima = np.intersect1d(argmaxima2, argmaxima)

        # Now find which intersection points are "best"
        # TODO: have user specify metric that they care about
        # https://en.wikipedia.org/wiki/Confusion_matrix
        if True:
            # Use area under curves to determine the probability density of tp,
            # fp, tn, fn at each candidate threshold.
            tp_area = get_right_area(tp_curve, xdata, argmaxima)
            fp_area = get_right_area(tn_curve, xdata, argmaxima)
            tn_area = get_left_area(tn_curve, xdata, argmaxima)
            fn_area = get_left_area(tp_curve, xdata, argmaxima)  # NOQA
            # Choose the location of intersection that performs best on some test
            # statistic. (Positive likelihood ratio)
            lr_pos_ = tp_area / fp_area
            lr_neg_ = fn_area / tn_area
            # Accuracy is (tp + tn) / total
            acc = (tp_area + tn_area) / 2
            #print('lr_neg = %r' % (lr_neg,))
            #print('lr_pos = %r' % (lr_pos,))
            # Normalize likelihood into range 0 to 1
            pos_norm = max(1, vt.safe_max(lr_pos_, fill=1, finite=True, nans=False))
            neg_norm = max(1, vt.safe_max(lr_neg_, fill=1, finite=True, nans=False))
            lr_pos = lr_pos_ / pos_norm  # NOQA
            lr_neg = lr_neg_ / neg_norm  # NOQA

            #chosen_metric = lr_pos
            chosen_metric = acc
            # Invalidate impossible values
            isvalid = np.isfinite(chosen_metric)
            if np.any(isvalid):
                valid_argmaxima = argmaxima[isvalid]
                chosen_metric = chosen_metric[isvalid]
            else:
                valid_argmaxima = argmaxima

            # Invalidate values based on "reasonable" heuristics
            #reasonable_tp = tp_area > .1
            #reasonable_fp = tn_area > .1
            #reasonable_flags = np.logical_and(reasonable_tp, reasonable_fp)
            #if np.any(reasonable_flags):
            #    lr_pos = lr_pos.compress(reasonable_flags)
            #    lr_neg = lr_neg.compress(reasonable_flags)

            # Choose a finite argmax
            sortx = chosen_metric.argsort()[::-1]
            closeness_argmax = valid_argmaxima[sortx[0]]
            #maxpos = valid_argmaxima[lr_pos.argmax()]
            #print('lr_pos.argmax = %r' % (lr_pos.argmax,))
            # Hack for infinity and nans. bring thems out of the 0 and 1 range, but only by a bit.
            #lr_pos[np.isnan(lr_pos)] = -.1
            #lr_neg[np.isnan(lr_neg)] = -.1
            #lr_pos[np.isinf(lr_pos)] = 1.1
            #lr_neg[np.isinf(lr_neg)] = 1.1
        else:
            closeness_argmax = closeness.argmax()

        if closeness_argmax == len(closeness) - 1:
            y_submax = closeness[-2:-1]
            x_submax = xdata[-2:-1]
        elif closeness_argmax == 0:
            y_submax = closeness[0:1]
            x_submax = xdata[0:1]
        else:
            # argmaxima, hist_, centers = maxpos, closeness, xdata
            x_submax, y_submax = vt.interpolate_submaxima(np.array([closeness_argmax]), closeness, xdata)
        score_thresh = x_submax[0]
        if ut.get_argflag('--debug-scorethresh') and not getattr(encoder, 'block', False):
            encoder.block = True
            ut.exec_func_doctest(encoder.learn_threshold2,
                                 start_sentinal='import plottool_ibeis as pt',
                                 end_sentinal='pt.show_if_requested()')
            encoder.block = False
        return score_thresh

    def learn_threshold(encoder, verbose=False, **thresh_kw):
        """
        Learns cutoff threshold that achieves the target confusion metric
        Typically a desired false positive rate (recall) is specified
        """
        import vtool_ibeis as vt
        # select a cutoff threshold
        #import sklearn.metrics
        if len(thresh_kw) > 0:
            _thresh_kw = ut.map_dict_keys(lambda x: x.replace('target_', ''), thresh_kw)
        else:
            _thresh_kw = encoder.thresh_kw
        # Select threshold that gives target confusion
        _selected_items = [item for item in _thresh_kw.items()
                           if item[1] is not None]
        assert len(_selected_items) == 1, (
            'Can only specify one desired confusion metric')
        # choose how to optimize the threshold
        metric, value = _selected_items[0]
        # Get classifier confusions (maybe dont need probs here)
        X, y, attrs = encoder.get_support()
        probs = encoder.normalize_scores(X)

        if False:
            confusions_score = vt.ConfusionMetrics().fit(
                -X, y, verbose=verbose)

            confusions_prob = vt.ConfusionMetrics().fit(
                probs, y, verbose=verbose)

            _score_thresh = confusions_score.get_thresh_at_metric(metric, value)
            _prob_thresh = confusions_prob.get_thresh_at_metric(metric, value)
            _inv_score = encoder.inverse_normalize(_prob_thresh)
            _inv_prob = encoder.normalize_scores(-_score_thresh)

            _inv_prob - _prob_thresh
            _inv_score - (-_score_thresh)

        confusions = vt.ConfusionMetrics().fit(
            probs, y, verbose=verbose)

        prob_thresh = confusions.get_thresh_at_metric(metric, value)

        #target_value = confusions.get_metric_at_thresh(metric, prob_thresh)
        #check_thresh = confusions.get_thresh_at_metric(metric, target_value)
        score_thresh = encoder.inverse_normalize(prob_thresh)
        if verbose:
            print('[scorenorm] Learning threshold to achieve %s=%.5f' % (
                metric.upper(), value,))
            if encoder.learned_thresh is not None:
                print('[scorenorm]   * learned_thresh = %.5f' % (
                    encoder.learned_thresh,))
            else:
                print('[scorenorm]   * learned_thresh = %r' % (encoder.learned_thresh,))
            print('[scorenorm]   * score_thresh = %.5f' % (
                score_thresh,))
            if metric == 'tpr':
                print('[scorenorm]   * fpr = %r' % (
                    confusions.get_fpr_at_recall(value),))
            elif metric == 'fpr':
                print('[scorenorm]   * tpr = %r' % (
                    confusions.get_recall_at_fpr(value),))

        # TODO: maybe do not change state?
        encoder.learned_thresh = prob_thresh
        return score_thresh

    def inverse_normalize(encoder, probs):
        inverse_interp = scipy.interpolate.interp1d(
            encoder.p_tp_given_score, encoder.score_domain, kind='linear',
            copy=False, assume_sorted=False)
        scores = inverse_interp(probs)
        return scores

    def normalize_scores(encoder, X):
        is_iterable = ut.isiterable(X)
        if not is_iterable:
            X = np.array([X])
        prob = normalize_scores(
            encoder.score_domain, encoder.p_tp_given_score,
            X, interp_fn=encoder.interp_fn)
        if not is_iterable:
            prob = prob[0]
        return prob

    def predict(encoder, X):
        """ Predict true or false of ``X``. """
        prob = encoder.normalize_scores(X)
        pred = prob > encoder.learned_thresh
        return pred

    def get_accuracy(encoder, X, y):
        pred = encoder.predict(X)
        is_correct = pred == y
        accuracy = (is_correct).mean()
        return accuracy

    def get_error_indicies(encoder, X, y):
        r"""
        Returns the indicies of the most difficult type I and type II errors.

        Example:
            >>> # DISABLE_DOCTEST
            >>> from vtool_ibeis.score_normalization import *  # NOQA
            >>> encoder, X, y = testdata_score_normalier()
            >>> (fp_indicies, fn_indicies) = encoder.get_error_indicies(X, y)
            >>> fp_X = X.take(fp_indicies)[0:3]
            >>> fn_X = X.take(fn_indicies)[0:3]
            >>> result =    'fp_X = ' + ub.repr2(fp_X)
            >>> result += '\nfn_X = ' + ub.repr2(fn_X)
            >>> print(result)
            fp_X = np.array([ 6.196,  5.912,  5.804])
            fn_X = np.array([ 3.947,  4.277,  4.43 ])
        """
        prob = encoder.normalize_scores(X)
        pred = prob > encoder.learned_thresh
        is_correct = pred == y
        # Find misspredictions
        is_error = np.logical_not(is_correct)
        # get indexes of misspredictions
        error_indicies = np.where(is_error)[0]
        # Separate by Type I and Type II error
        error_y = y.take(error_indicies)
        fp_indicies_ = error_indicies.compress(np.logical_not(error_y))
        fn_indicies_ = error_indicies.compress(error_y)
        # Sort errors by difficulty
        fp_sortx = prob.take(fp_indicies_).argsort()[::-1]
        fn_sortx = prob.take(fn_indicies_).argsort()
        fp_indicies = fp_indicies_.take(fp_sortx)
        fn_indicies = fn_indicies_.take(fn_sortx)
        return fp_indicies, fn_indicies

    def get_correct_indices(encoder, X, y):
        r"""
        Args:
            X (ndarray):  data
            y (ndarray):  labels

        Returns:
            tuple: (fp_indicies, fn_indicies)

        CommandLine:
            python -m vtool_ibeis.score_normalization --test-get_correct_indices

        Example:
            >>> # DISABLE_DOCTEST
            >>> from vtool_ibeis.score_normalization import *  # NOQA
            >>> encoder, X, y = testdata_score_normalier()
            >>> (tp_indicies, tn_indicies) = encoder.get_correct_indices(X, y)
            >>> tp_X = X.take(tp_indicies)[0:3]
            >>> tn_X = X.take(tn_indicies)[0:3]
            >>> result =    'tp_X = ' + ub.repr2(tp_X)
            >>> result += '\ntn_X = ' + ub.repr2(tn_X)
            >>> print(result)
            tp_X = np.array([ 8.883,  8.77 ,  8.759])
            tn_X = np.array([ 0.727,  0.76 ,  0.841])
        """
        prob = encoder.normalize_scores(X)
        pred = prob > encoder.learned_thresh
        is_correct = pred == y
        # get indexes of misspredictions
        correct_indicies = np.where(is_correct)[0]
        # Separate by Type I and Type II error
        correct_y = y.take(correct_indicies)
        tp_indicies_ = correct_indicies.compress(correct_y)
        tn_indicies_ = correct_indicies.compress(np.logical_not(correct_y))
        # Sort by most correct cases
        tp_sortx = prob.take(tp_indicies_).argsort()[::-1]
        tn_sortx = prob.take(tn_indicies_).argsort()
        tp_indicies = tp_indicies_.take(tp_sortx)
        tn_indicies = tn_indicies_.take(tn_sortx)
        return tp_indicies, tn_indicies

    def get_confusion_indicies(encoder, X, y):
        """
        combination of get_correct_indices and get_error_indicies
        """
        prob = encoder.normalize_scores(X)
        pred = prob > encoder.learned_thresh
        # Find correct and misspredictions
        is_correct = pred == y
        is_error = np.logical_not(is_correct)
        correct_indicies = np.where(is_correct)[0]
        error_indicies = np.where(is_error)[0]
        # Separate by Correct and Type I and Type II error
        error_y = y.take(error_indicies)
        correct_y = y.take(correct_indicies)
        fp_indicies_ = error_indicies.compress(np.logical_not(error_y))
        fn_indicies_ = error_indicies.compress(error_y)
        tp_indicies_ = correct_indicies.compress(correct_y)
        tn_indicies_ = correct_indicies.compress(np.logical_not(correct_y))
        # Sort errors by difficulty and other cases by correctness
        fp_sortx = prob.take(fp_indicies_).argsort()[::-1]
        fn_sortx = prob.take(fn_indicies_).argsort()
        tp_sortx = prob.take(tp_indicies_).argsort()[::-1]
        tn_sortx = prob.take(tn_indicies_).argsort()
        indicies = ut.DynStruct()
        indicies.tp = tp_indicies_.take(tp_sortx)
        indicies.tn = tn_indicies_.take(tn_sortx)
        indicies.fp = fp_indicies_.take(fp_sortx)
        indicies.fn = fn_indicies_.take(fn_sortx)
        return indicies

    def visualize(encoder, **kwargs):
        r"""
        shows details about the score normalizer

        Kwargs:
            fnum
            figtitle
            with_hist
            interactive
            with_scores
            with_roc
            with_precision_recall

        CommandLine:
            python -m vtool_ibeis.score_normalization --exec-ScoreNormalizer.visualize:0 --show
            python -m vtool_ibeis.score_normalization --exec-ScoreNormalizer.visualize:1 --show

        Example0:
            >>> # UNSTABLE_DOCTEST
            >>> from vtool_ibeis.score_normalization import *  # NOQA
            >>> import vtool_ibeis as vt
            >>> encoder = ScoreNormalizer()
            >>> X, y = vt.demodata.testdata_binary_scores()
            >>> encoder.fit(X, y)
            >>> kwargs = dict(
            >>>     with_pr=True, interactive=True, with_roc=True,
            >>>     with_hist=True)
            >>> encoder.visualize(**kwargs)
            >>> ut.show_if_requested()

        Example1:
            >>> # UNSTABLE_DOCTEST
            >>> from vtool_ibeis.score_normalization import *  # NOQA
            >>> import vtool_ibeis as vt
            >>> encoder = ScoreNormalizer()
            >>> X, y = vt.demodata.testdata_binary_scores()
            >>> encoder.fit(X, y)
            >>> kwargs = dict(
            >>>     with_pr=True, interactive=True, with_roc=True, with_hist=True,
            >>>     with_scores=False, with_prebayes=False, with_postbayes=False)
            >>> encoder.visualize(target_tpr=.95, **kwargs)
            >>> ut.show_if_requested()
        """
        #import plottool_ibeis as pt
        default_kw = dict(
            with_scores=False,
            with_roc=True,
            #with_roc=False,
            #with_precision_recall=True,
            with_precision_recall=False,
            #with_hist=False,
            with_hist=True,
            fnum=None,
            figtitle=None,
            #interactive=None,
            interactive=False,
            use_stems=None,
            attr_callback=None,
            with_prebayes=True,
            with_postbayes=True,
            score_range=None,
            bin_width=None,
            logscale=False,
        )
        alias_dict = {'with_pr': 'with_precision_recall'}
        # inspect_kw = ut.update_existing(default_kw, kwargs, alias_dict=alias_dict)
        inspect_kw = ut.update_dict(default_kw, kwargs, alias_dict=alias_dict)
        print('inspect_kw = %r' % (inspect_kw,))
        other_kw = ut.delete_dict_keys(kwargs.copy(), list(inspect_kw.keys()) +
                                       list(alias_dict.keys()))

        score_thresh, prob_thresh = encoder._hack_vizlearn(**other_kw)

        tup = encoder.get_partitioned_support()
        tp_support, tn_support, part_attrs = tup
        inter = inspect_pdfs(
            tn_support, tp_support,
            encoder.score_domain, encoder.p_tp_given_score,
            encoder.p_tn_given_score, encoder.p_score_given_tp,
            encoder.p_score_given_tn, encoder.p_score,
            prob_thresh=prob_thresh,
            score_thresh=score_thresh, part_attrs=part_attrs,
            thresh_kw=encoder.thresh_kw,
            **inspect_kw)
        return inter


def partition_scores(X, y, attrs=None):
    """
    convinience helper to translate partitioned to unpartitioned data

    Args:
        tp_scores (ndarray):
        tn_scores (ndarray):
        attrs (dict): (default = None)

    Returns:
        tuple: (scores, labels, attrs)

    CommandLine:
        python -m vtool_ibeis.score_normalization --test-partition_scores

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.score_normalization import *  # NOQA
        >>> X = np.array([5, 6, 6, 7, 1, 2, 2])
        >>> attrs = {'qaid': np.array([21, 24, 25, 26, 11, 14, 15])}
        >>> y = np.array([1, 1, 1, 1, 0, 0, 0], dtype=bool)
        >>> tup = partition_scores(X, y, attrs)
        >>> resdict = ut.odict(zip(
        >>>     ['tp_scores', 'tn_scores', 'part_attrs'], tup))
        >>> result = ub.repr2(resdict, nobraces=True, with_dtype=False,
        >>>                      explicit=1, nl=2)
        >>> print(result)

        tp_scores=np.array([5, 6, 6, 7]),
        tn_scores=np.array([1, 2, 2]),
        part_attrs={
            False: {'qaid': np.array([11, 14, 15])},
            True: {'qaid': np.array([21, 24, 25, 26])},
        },
    """
    import vtool_ibeis as vt
    import operator
    # Make partitioning
    unique_labels, groupxs = vt.group_indices(y)
    _grouper = partial(vt.apply_grouping, groupxs=groupxs)
    # Group data
    X_parts = _grouper(X)
    # Group attributes
    _nested_attrs = ut.map_dict_vals(_grouper, attrs)
    def _getitem(a, b):
        return operator.getitem(b, a)
    part_attrs = {
        label: ut.map_dict_vals(partial(_getitem, lblx), _nested_attrs)
        for lblx, label in enumerate(unique_labels)
    }
    assert len(unique_labels) == 2, 'exepcted two groups'
    assert not unique_labels[0], 'expected true negatives to be first'
    assert unique_labels[1], 'expected true positives to be second'
    #
    tn_scores, tp_scores = X_parts
    return tp_scores, tn_scores, part_attrs


def flatten_scores(tp_scores, tn_scores, part_attrs=None):
    """
    convinience helper to translate partitioned to unpartitioned data

    Args:
        tp_scores (ndarray):
        tn_scores (ndarray):
        part_attrs (dict): (default = None)

    Returns:
        tuple: (scores, labels, attrs)

    CommandLine:
        python -m vtool_ibeis.score_normalization --test-flatten_scores

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.score_normalization import *  # NOQA
        >>> tp_scores = np.array([5, 6, 6, 7])
        >>> tn_scores = np.array([1, 2, 2])
        >>> part_attrs = {
        ...     1: {'qaid': [21, 24, 25, 26]},
        ...     0: {'qaid': [11, 14, 15]},
        ... }
        >>> tup = flatten_scores(
        ... tp_scores, tn_scores, part_attrs)
        >>> (X, y, attrs) = tup
        >>> y = y.astype(int)
        >>> resdict = ut.odict(zip(['X', 'y', 'attrs'], [X, y, attrs]))
        >>> result = ub.repr2(resdict, nobraces=True, with_dtype=False,
        >>>                      explicit=1, nl=1)
        >>> print(result)

        X=np.array([5, 6, 6, 7, 1, 2, 2]),
        y=np.array([1, 1, 1, 1, 0, 0, 0]),
        attrs={'qaid': np.array([21, 24, 25, 26, 11, 14, 15])},
    """
    scores = np.hstack([tp_scores, tn_scores])
    labels = np.zeros(scores.size, dtype=bool)
    labels[0:len(tp_scores)] = True
    if part_attrs is None:
        return scores, labels
    else:
        tp_attrs = part_attrs[1]
        tn_attrs = part_attrs[0]
        assert (tp_attrs is None) == (tn_attrs is None), (
            'must specify both or none')
        assert sorted(tp_attrs.keys()) == sorted(tn_attrs.keys()), (
            'dicts do not agree')
        # attrs = ut.dict_isect_combine(tp_attrs, tn_attrs, combine_op=np.append)
        from functools import partial
        combine_op = partial(np.append, axis=0)
        attrs = ut.dict_isect_combine(tp_attrs, tn_attrs, combine_op=combine_op)
        num_attrs = np.array(list(map(len, attrs.values())))
        assert np.all(num_attrs == len(scores)), (
            'num_attrs=%r must agree with data. len(scores)=%r' % (
                num_attrs, len(scores)))
        return scores, labels, attrs


def learn_score_normalization(tp_support, tn_support, gridsize=1024, adjust=8,
                              return_all=False, monotonize=True,
                              clip_factor=(ut.PHI + 1), verbose=False,
                              reverse=False, p_tp_method='eq'):
    r"""
    Takes collected data and applys parzen window density estimation and bayes rule.

    #True positive scores must be larger than true negative scores.
    FIXME: might be an issue with pdfs summing to 1 here.

    Args:
        tp_support (ndarray):
        tn_support (ndarray):
        gridsize       (int): default 512
        adjust         (int): default 8
        return_all     (bool): default False
        monotonize     (bool): default True
        clip_factor    (float): default phi ** 2

    Returns:
        tuple: (score_domain, p_tp_given_score, p_tn_given_score, p_score_given_tp, p_score_given_tn, p_score)

    CommandLine:
        python -m vtool_ibeis.score_normalization --test-learn_score_normalization

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.score_normalization import *  # NOQA
        >>> tp_support = np.linspace(100, 10000, 512)
        >>> tn_support = np.linspace(0, 120, 512)
        >>> gridsize = 1024
        >>> adjust = 8
        >>> return_all = False
        >>> monotonize = True
        >>> clip_factor = 2.6180339887499997
        >>> verbose = True
        >>> reverse = False
        >>> (score_domain, p_tp_given_score) = learn_score_normalization(tp_support, tn_support)
        >>> result = '%.2f' % (np.diff(p_tp_given_score).sum())
        >>> print(result)
        0.99
    """
    import vtool_ibeis as vt
    if verbose:
        print('[scorenorm] Learning normalization pdf')
        print('[scorenorm] * tp_support.shape=%r' % (tp_support.shape,))
        print('[scorenorm] * tn_support.shape=%r' % (tn_support.shape,))
        print('[scorenorm] * estimating true positive pdf, ')
        print('[scorenorm] * monotonize = %r' % (monotonize,))
        print('stats.tp_support = ' + ut.get_stats_str(tp_support, use_nan=True))
        print('stats.tn_support = ' + ut.get_stats_str(tn_support, use_nan=True))
        next_ = ut.next_counter(1)
        total = 8
    # import utool
    # utool.embed()
    # Find good score domain range
    if True:
        min_score, max_score = find_clip_range(tp_support, tn_support, clip_factor, reverse)
    else:
        min_score = min(tp_support.min(), tn_support.min())
        max_score = min(tp_support.max(), tn_support.max())
    score_domain = np.linspace(min_score, max_score, gridsize)
    # Estimate true positive/negative density
    if verbose:
        print('[scorenorm] %d/%d estimating true negative pdf' % (next_(), total))
    score_tp_pdf = vt.estimate_pdf(tp_support, gridsize=gridsize, adjust=adjust)
    #assert score_tp_pdf.bw != 0, 'error bandwidth estimated to be 0'
    if verbose:
        print('[scorenorm] %d/%d estimating true negative pdf' % (next_(), total))
    score_tn_pdf = vt.estimate_pdf(tn_support, gridsize=gridsize, adjust=adjust)
    #assert score_tn_pdf.bw != 0, 'error bandwidth estimated to be 0'
    if verbose:
        print('[scorenorm] %d/%d estimating score domain' % (next_(), total))
    # Evaluate true negative density
    if verbose:
        print('[scorenorm] %d/%d evaluating tp density' % (next_(), total))
    p_score_given_tp = score_tp_pdf.evaluate(score_domain)
    if verbose:
        print('[scorenorm] %d/%d evaluating tn density' % (next_(), total))
    p_score_given_tn = score_tn_pdf.evaluate(score_domain)

    # Not sure why the pdfs returned from statsmodels dont integrate to 1

    if verbose:
        print('[sn.pre]stats.score_domain = ' + ut.get_stats_str(score_domain, use_nan=True, precision=5))
        print('[sn.pre]stats:p_score_given_tn = ' + ut.get_stats_str(p_score_given_tn, use_nan=True, precision=5))
        print('[sn.pre]stats:p_score_given_tp = ' + ut.get_stats_str(p_score_given_tp, use_nan=True, precision=5))
        #print('stats.tn_support = ' + ut.get_stats_str(tn_support, use_nan=True))

    problems = []

    try:
        assert not np.any(np.isnan(p_score_given_tp)), ('Need more positive support')
    except AssertionError as ex:  # NOQA
        print('[sn.pre]stats:tpsupport = ' + ut.get_stats_str(score_tp_pdf.support, use_nan=True, precision=5))
        problems += [str(ex)]

    try:
        assert not np.any(np.isnan(p_score_given_tn)), ('Need more negative support')
    except AssertionError as ex:  # NOQA
        print('[sn.pre]stats:tnsupport = ' + ut.get_stats_str(score_tn_pdf.support, use_nan=True, precision=5))
        problems += [str(ex)]

    if problems:
        raise AssertionError(', '.join(problems))

    if True:
        # Make sure we still have probability functions
        area_tp = np.trapz(p_score_given_tp, score_domain)
        area_tn = np.trapz(p_score_given_tn, score_domain)
        if verbose:
            print('pre.area_tp = %r' % (area_tp,))
            print('pre.area_tn = %r' % (area_tn,))

        # normalize to ensure
        p_score_given_tp = p_score_given_tp / area_tp
        p_score_given_tn = p_score_given_tn / area_tn

        area_tp = np.trapz(p_score_given_tp, score_domain)
        area_tn = np.trapz(p_score_given_tn, score_domain)
        #if ut.DEBUG2:
        if verbose:
            print('norm.area_tp = %r' % (area_tp,))
            print('norm.area_tn = %r' % (area_tn,))

        assert np.isclose(area_tp, 1.0), area_tp
        assert np.isclose(area_tn, 1.0), area_tn

    if verbose:
        print('[scorenorm] %d/%d evaluating posterior probabilities' % (next_(), total))
        print('p_tp_method = %r' % (p_tp_method,))

    # For inbalanced data there are several methods we might want to use to
    # calculate p_tp.  not always going to be equal probability of true and
    # positive cases
    if p_tp_method == 'eq':
        p_tp = .5
    elif p_tp_method == 'ratio':
        p_tp = len(tp_support) / (len(tp_support) + len(tn_support))
    else:
        raise NotImplementedError('p_tp_method = %r' % (p_tp_method,))
    # Average to get probablity of any score
    p_score = (np.array(p_score_given_tp) + np.array(p_score_given_tn)) / 2.0
    # Apply bayes
    p_tp_given_score = ut.bayes_rule(p_score_given_tp, p_tp, p_score)
    if ut.DEBUG2:
        assert np.isclose(np.trapz(p_score, score_domain), 1.0)
        assert np.isclose(np.trapz(p_score, p_tp_given_score), 1.0)
    if np.any(np.isnan(p_tp_given_score)):
        p_tp_given_score = vt.interpolate_nans(p_tp_given_score)
    if verbose:
        # np.trapz(p_tp_given_score / np.trapz(p_tp_given_score, score_domain), score_domain)
        print('stats:p_score_given_tn = ' + ut.get_stats_str(p_score_given_tn, newlines=0, use_nan=True, precision=5))
        print('stats:p_score_given_tp = ' + ut.get_stats_str(p_score_given_tp, newlines=0, use_nan=True, precision=5))
        print('stats:p_score = ' + ut.get_stats_str(p_score, newlines=0, use_nan=True, precision=5))
        print('stats:p_tp_given_score = ' + ut.get_stats_str(p_tp_given_score, newlines=0, use_nan=True, precision=5))
    if monotonize:
        if reverse:
            if verbose:
                print('[scorenorm] %d/%d monotonize decreasing' % (next_(), total))
            p_tp_given_score = vt.ensure_monotone_strictly_decreasing(
                p_tp_given_score, left_endpoint=1.0, right_endpoint=0.0)
        else:
            if verbose:
                print('[scorenorm] %d/%d monotonize increasing' % (next_(), total))
            #if False:
            #    flags = ~np.isnan(p_tp_given_score)
            #    pt.plot(score_domain[flags], p_tp_given_score[flags])
            #    pt.plot(score_domain, p_tp_given_score)
            p_tp_given_score = vt.ensure_monotone_strictly_increasing(
                p_tp_given_score, left_endpoint=0.0, right_endpoint=1.0)
    if return_all:
        if verbose:
            print('[scorenorm] %d/%d returning all' % (next_(), total))
        p_tn_given_score = 1 - p_tp_given_score
        return (score_domain, p_tp_given_score, p_tn_given_score,
                p_score_given_tp, p_score_given_tn, p_score)
    else:
        if verbose:
            print('[scorenorm] %d/%d returning minimum' % (next_(), total))
        return (score_domain, p_tp_given_score)


def find_clip_range(tp_support, tn_support, clip_factor=ut.PHI + 1, reverse=None):
    """
    TODO: generalize to arbitrary domains (not just 0->inf)

    Finds score to clip true positives past. This is useful when the highest
    true positive scores can be much larger than the highest true negative
    score.

    Args:
        tp_support (ndarray):
        tn_support (ndarray):
        clip_factor (float): factor of the true negative domain to search for true positives

    Returns:
        tuple: min_score, max_score

    CommandLine:
        python -m vtool_ibeis.score_normalization --test-find_clip_range

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.score_normalization import *  # NOQA
        >>> tp_support = np.array([100, 200, 50000])
        >>> tn_support = np.array([10, 30, 110])
        >>> clip_factor = ut.PHI + 1
        >>> min_score, max_score = find_clip_range(tp_support, tn_support,  clip_factor)
        >>> result = '%.4f, %.4f' % ((min_score, max_score))
        >>> print(result)
        10.0000, 287.9837
    """
    if reverse is None:
        mean_tp_score = tp_support.mean()
        mean_tn_score = tn_support.mean()
        reverse = mean_tp_score < mean_tn_score

    if not reverse:
        # Normal case where higher scores is better
        high_scores = tp_support
        low_scores  = tn_support
    else:
        high_scores = tn_support
        low_scores  = tp_support

    max_high_score = high_scores.max()
    max_low_score  = low_scores.max()
    min_high_score = high_scores.min()
    min_low_score  = low_scores.min()
    abs_max_score = max(max_high_score, max_low_score)
    abs_min_score = min(min_high_score, min_low_score)

    if clip_factor is None:
        min_score = abs_min_score
        max_score = abs_max_score
        return min_score, max_score

    # FIXME: allow for true positive scores to be low, or not bounded at 0

    # Do not clip if true negatives can score higher than true positives
    if max_low_score < max_high_score:
        #overshoot_factor = (max_high_score - abs_min_score) / (max_low_score - abs_min_score)
        overshoot_factor = max_high_score / max_low_score
        if overshoot_factor > clip_factor:
            max_score = max_low_score * clip_factor
        else:
            max_score = max_high_score
    min_score = abs_min_score
    #if min_low_score < min_high_score:
    #    overshoot_factor = min_low_score / min_high_score
    #    if overshoot_factor > clip_factor:
    #        min_score = min_high_score * clip_factor
    #    else:
    #        min_score = min_low_score
    return min_score, max_score


def normalize_scores(score_domain, p_tp_given_score, scores, interp_fn=None):
    """
    Adjusts a raw scores to a probabilities based on a learned normalizer

    Args:
        score_domain (ndarray): input score domain
        p_tp_given_score (ndarray): learned probability mapping
        scores (ndarray): raw scores

    Returns:
        ndarray: probabilities

    CommandLine:
        python -m vtool_ibeis.score_normalization --test-normalize_scores

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.score_normalization import *  # NOQA
        >>> score_domain = np.linspace(0, 10, 10)
        >>> p_tp_given_score = (score_domain ** 2) / (score_domain.max() ** 2)
        >>> scores = np.array([-1, 0.0, 0.01, 2.3, 8.0, 9.99, 10.0, 10.1, 11.1])
        >>> prob = normalize_scores(score_domain, p_tp_given_score, scores)
        >>> #np.set_printoptions(suppress=True)
        >>> result = ub.repr2(prob, precision=2, suppress_small=True)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.plot2(score_domain, p_tp_given_score, 'r-x', equal_aspect=False, label='learned probability')
        >>> pt.plot2(scores, prob, 'yo', equal_aspect=False, title='Normalized scores', pad=.2, label='query points')
        >>> pt.legend('upper left')
        >>> ut.show_if_requested()
        np.array([ 0.  ,  0.  ,  0.  ,  0.05,  0.64,  1.  ,  1.  ,  1.  ,  1.  ], dtype=np.float64)
    """
    #prob = np.zeros(len(scores))
    prob = np.zeros(len(scores))
    #prob = np.full(len(scores), np.nan)
    is_nan  = np.isnan(scores)
    # Check score domain bounds
    is_low  = scores < score_domain[0]
    is_high = scores > score_domain[-1]
    is_inbounds = np.logical_not(np.logical_or.reduce((is_nan, is_low, is_high)))
    # interpolate scores in the learned domain
    # we are garuenteed to have inbounds nonzero elements here
    if True:
        if interp_fn is None:
            # TODO build custom interpolator with correct bound checks
            interp_fn = scipy.interpolate.interp1d(score_domain, p_tp_given_score,
                                                   kind='linear', copy=False,
                                                   assume_sorted=True)
        prob[is_inbounds] = interp_fn(scores[is_inbounds])
    else:
        flags = score_domain <= scores[is_inbounds][:, None]
        left_indicies = np.array([np.nonzero(row)[0][-1] for row in flags])
        prob[is_inbounds] = p_tp_given_score[left_indicies]
    # currently just taking the min
    # fill in other values
    assert not np.any(is_nan), 'cannot normalize nan values'
    #if any(is_nan):
    #    # handle nans
    #    raise AssertionError('user normalize score list')
    #    prob[np.isnan(score_domain)] = -1.0
    # clip low scores at 0
    prob[is_low] = 0
    # clip high scores by between max probability and one
    prob[is_high] = (p_tp_given_score[-1] + 1.0) / 2.0
    return prob


# DEBUGGING FUNCTIONS


def test_score_normalization(tp_support, tn_support, with_scores=True,
                             verbose=True, with_roc=True,
                             with_precision_recall=False, figtitle=None,
                             normkw_varydict=None):
    """
    Gives an overview of how well threshold can be learned from raw scores.

    DEPRICATE

    CommandLine:
        python -m vtool_ibeis.score_normalization --test-test_score_normalization --show

    CommandLine:
        xdoctest -m ~/code/vtool_ibeis/vtool_ibeis/score_normalization.py test_score_normalization

    Ignore:
        >>> # GUI_DOCTEST
        >>> # Shows how score normalization works with gaussian noise
        >>> from vtool_ibeis.score_normalization import *  # NOQA
        >>> verbose = True
        >>> randstate = np.random.RandomState(seed=0)
        >>> # Get a training sample
        >>> tp_support = randstate.normal(loc=6.5, size=(256,))
        >>> tn_support = randstate.normal(loc=3.5, size=(256,))
        >>> # xdoctest: +REQUIRES(module:plottool_ibeis)
        >>> test_score_normalization(tp_support, tn_support, verbose=verbose)
        >>> ut.show_if_requested()

    """
    import plottool_ibeis as pt  # NOQA

    # Print raw score statistics
    print('tp_support')
    print(ut.repr4(ut.get_stats(tp_support)))
    print('tn_support')
    print(ut.repr4(ut.get_stats(tn_support)))

    # Test (potentially multiple) normalizing configurations
    if normkw_varydict is None:
        normkw_varydict = {
            'monotonize': [False],  # [True, False],
            #'adjust': [1, 4, 8],
            'adjust': [1],
            #'adjust': [8],
        }
    normkw_list = ut.util_dict.all_dict_combinations(normkw_varydict)

    if len(normkw_list) > 32:
        raise AssertionError('Too many plots to test!')

    for normkw in normkw_list:
        # Learn the appropriate normalization
        tup = learn_score_normalization(tp_support, tn_support,
                                         return_all=True, verbose=verbose,
                                         **normkw)
        (score_domain,
         p_tp_given_score, p_tn_given_score,
         p_score_given_tp, p_score_given_tn,
         p_score) = tup

        if verbose:
            print('plotting pdfs')
        fnum = pt.next_fnum()

        inspect_pdfs(tn_support, tp_support, score_domain, p_tp_given_score,
                     p_tn_given_score, p_score_given_tp, p_score_given_tn,
                     p_score, with_scores=with_scores, with_roc=with_roc,
                     with_precision_recall=with_precision_recall, fnum=fnum)

        if figtitle is not None:
            pt.set_figtitle(figtitle)
        else:
            pt.set_figtitle('ScoreNorm test' + ub.repr2(normkw, newlines=False))

    locals_ = locals()
    return locals_


# --------
# Plotting
# --------


def plot_prebayes_pdf(score_domain, p_score_given_tn, p_score_given_tp, p_score,
                      cfgstr='', fnum=None, pnum=(1, 1, 1), **kwargs):
    import plottool_ibeis as pt
    if fnum is None:
        fnum = pt.next_fnum()
    true_color = pt.TRUE_BLUE  # pt.TRUE_GREEN
    false_color = pt.FALSE_RED
    #unknown_color = pt.UNKNOWN_PURP
    unknown_color = pt.PURPLE2
    #unknown_color = pt.GRAY

    pt.plots.plot_probabilities(
        (p_score_given_tn,  p_score_given_tp, p_score),
        ('p(score | tn)', 'p(score | tp)', 'p(score)'),
        prob_colors=(false_color, true_color, unknown_color),
        #figtitle='pre_bayes pdf score',
        figtitle='p(score | truth)',
        xdata=score_domain,
        fnum=fnum,
        pnum=pnum, **kwargs)


def plot_postbayes_pdf(score_domain, p_tn_given_score, p_tp_given_score,
                       score_thresh=None, prob_thresh=None, cfgstr='',
                       fnum=None, pnum=(1, 1, 1)):
    import plottool_ibeis as pt
    if fnum is None:
        fnum = pt.next_fnum()
    true_color = pt.TRUE_BLUE  # pt.TRUE_GREEN
    false_color = pt.FALSE_RED

    pt.plots.plot_probabilities(
        (p_tn_given_score, p_tp_given_score),
        ('p(tn | score)', 'p(tp | score)'),
        prob_colors=(false_color, true_color,),
        #figtitle='post_bayes pdf score ' + cfgstr,
        figtitle='p(truth | score)' + cfgstr,
        xdata=score_domain, fnum=fnum, pnum=pnum,
        score_thresh=score_thresh,
        prob_thresh=prob_thresh)


def inspect_pdfs(tn_support, tp_support,
                 score_domain, p_tp_given_score,
                 p_tn_given_score, p_score_given_tp, p_score_given_tn, p_score,
                 prob_thresh=None, score_thresh=None, with_scores=False,
                 with_roc=False, with_precision_recall=False, with_hist=False,
                 fnum=None, figtitle=None, interactive=None, use_stems=None,
                 part_attrs=None, thresh_kw=None, attr_callback=None,
                 with_prebayes=True, with_postbayes=True, score_range=None,
                 **kwargs):
    r"""
    Shows plots of learned thresholds

    CommandLine:
        python -m vtool_ibeis.score_normalization --test-ScoreNormalizer --show
        python -m vtool_ibeis.score_normalization --exec-ScoreNormalizer.visualize --show
    """
    import plottool_ibeis as pt  # NOQA
    from plottool_ibeis.interactions import ExpandableInteraction
    from plottool_ibeis.abstract_interaction import AbstractInteraction
    import vtool_ibeis as vt
    import plottool_ibeis as pt  # NOQA

    if fnum is None:
        fnum = pt.next_fnum()

    with_normscore = with_scores
    #with_prebayes = True
    #with_postbayes = True

    nSubplots = (with_normscore + with_prebayes + with_postbayes +
                 with_scores + with_roc + with_precision_recall +
                 with_hist)
    if nSubplots == 0:
        raise ValueError('Must choose at least one subplot')
    nRows, nCols = pt.get_square_row_cols(nSubplots)
    _pnumiter = pt.make_pnum_nextgen(nRows=nRows, nCols=nCols,
                                     nSubplots=nSubplots)

    #print('Always interactive even if: interactive = %r' % (interactive,))
    # Make a plottool_ibeis interaction
    inter = ExpandableInteraction(fnum, _pnumiter)

    scores = np.hstack([tn_support, tp_support])
    labels = np.array([False] * len(tn_support) + [True] * len(tp_support))

    # probs = encoder.normalize_scores(scores)
    probs = normalize_scores(score_domain, p_tp_given_score, scores)

    confusions = vt.ConfusionMetrics().fit(
        probs, labels)
    # Hack change confusion prob thresholds to score thresholds
    if False:
        # Fixme: assume sorted
        inverse_interp = scipy.interpolate.interp1d(
            p_tp_given_score, score_domain, kind='linear',
            copy=False, assume_sorted=False)
        confusions._orig_thresholds = confusions.thresholds
        confusions.thresholds = inverse_interp(confusions.thresholds)
        confusions._hackscores = scores

    true_color = pt.TRUE_BLUE  # pt.TRUE_GREEN
    false_color = pt.FALSE_RED

    support_kw = dict(
        score_lbls=kwargs.get('score_lbls', ('trueneg', 'truepos')),
        score_colors=(false_color, true_color),
        logscale=kwargs.get('logscale', False),
    )
    support_sort_kw = dict(
        score_markers=['^', 'v'],
        markersizes=[5, 5],
        use_stems=use_stems,
        **support_kw
    )

    class SortedScoreSupportInteraction(AbstractInteraction):
        def __init__(data_pdf, **kwargs):
            super(SortedScoreSupportInteraction, data_pdf).__init__(**kwargs)
            data_pdf.tn_support = tn_support
            data_pdf.tp_support = tp_support
            data_pdf.part_attrs = part_attrs
            data_pdf.attr_callback = attr_callback
            data_pdf.sel_mode = 'tn'

        def toggle_mode(data_pdf):
            if data_pdf.sel_mode == 'tn':
                data_pdf.sel_mode = 'tp'
            else:
                data_pdf.sel_mode = 'tn'
            print('TOGGLE data_pdf.sel_mode = %r' % (data_pdf.sel_mode,))

        @staticmethod
        def static_plot(fnum, pnum):
            pt.plots.plot_sorted_scores(
                (tn_support, tp_support),
                fnum=fnum, pnum=pnum,
                score_label='score',
                thresh=score_thresh,
                **support_sort_kw
            )

        def plot(data_pdf, fnum, pnum):
            data_pdf.static_plot(fnum, pnum)

        def on_key_press(data_pdf, event):
            #print('event = %r' % (event,))
            #print('event.key = %r' % (event.key,))
            if event.key == 't':
                data_pdf.toggle_mode()

        def on_click_inside(data_pdf, event, ex):
            import vtool_ibeis as vt
            #ax = event.inaxes
            #for l in ax.get_lines():
            #    print(l.get_label())
            tp_index, tp_dist = vt.closest_point(event.ydata, data_pdf.tp_support[:, None])
            tn_index, tn_dist = vt.closest_point(event.ydata, data_pdf.tn_support[:, None])
            print('closest tp_index = %r, %r' % (tp_index, tp_dist))
            print('closest tn_index = %r, %r' % (tn_index, tn_dist))
            SEL_TP = data_pdf.sel_mode == 'tp'
            print('data_pdf.sel_mode = %r' % (data_pdf.sel_mode,))
            if SEL_TP:
                tp_attrs = data_pdf.part_attrs[True]
                if len(tp_attrs) == 0:
                    print('No positive attrs')
                subattrs = ut.get_dict_column(tp_attrs, tp_index)
            else:
                tn_attrs = data_pdf.part_attrs[False]
                if len(tn_attrs) == 0:
                    print('No negative attrs')
                subattrs = ut.get_dict_column(tn_attrs, tn_index)
            print('subattrs = %r' % (subattrs,))
            if data_pdf.attr_callback is not None:
                print('Executing callback')
                data_pdf.attr_callback(**subattrs)
            #dists = vt.L1(event.ydata, data_pdf.tp_support[:, None])
            #index = dists.argsort()[0]
            #event.xdata
            # Find the nearest label
            pass

    #target_tpr = None
    target_tpr = confusions.get_metric_at_thresh('tpr', prob_thresh)
    #print('target_tpr = %r' % (target_tpr,))
    ROCInteraction = vt.interact_roc_factory(confusions, target_tpr,
                                             show_operating_point=True)

    def _score_support_hist(fnum, pnum):
        overlay_score_domain = None
        score_thresh_ = None
        if kwargs.get('histoverlay', True):
            overlay_score_domain = score_domain
            score_thresh_ = score_thresh

        print('support_kw = %r' % (support_kw,))

        pt.plot_score_histograms(
            (tn_support, tp_support),
            score_thresh=score_thresh_,
            score_label='score',
            fnum=fnum,
            pnum=pnum,
            bin_width=kwargs.get('bin_width', None),
            num_bins=kwargs.get('num_bins', None),
            overlay_prob_given_list=(p_score_given_tn, p_score_given_tp),
            overlay_score_domain=overlay_score_domain,
            xlim=score_range,
            histnorm=kwargs.get('histnorm', False),
            **support_kw)

    def _prob_support_hist(fnum, pnum):
        tp_probs = probs[labels]
        tn_probs = probs[np.logical_not(labels)]
        pt.plot_score_histograms(
            (tn_probs, tp_probs),
            score_label='prob',
            fnum=fnum,
            pnum=pnum,
            bin_width=kwargs.get('bin_width', None),
            num_bins=kwargs.get('num_bins', None),
            **support_kw)

    def _prob_support_sorted(fnum, pnum):
        tp_probs = probs[labels]
        tn_probs = probs[np.logical_not(labels)]
        pt.plots.plot_sorted_scores(
            (tn_probs, tp_probs),
            fnum=fnum, pnum=pnum,
            score_label='prob',
            thresh=prob_thresh,
            **support_sort_kw
        )
        #ax = pt.gca()
        #max_score = max(tn_support.max(), tp_support.max())
        #ax.set_ylim(-max_score, max_score)

    def _prebayes(fnum, pnum):
        plot_prebayes_pdf(
            score_domain, p_score_given_tn, p_score_given_tp, p_score,
            score_thresh=score_thresh,
            cfgstr='', fnum=fnum, pnum=pnum)

    def _postbayes(fnum, pnum):
        plot_postbayes_pdf(score_domain, p_tn_given_score, p_tp_given_score,
                           prob_thresh=prob_thresh,
                           score_thresh=score_thresh,
                           cfgstr='', fnum=fnum, pnum=pnum)

    def _precision_recall(fnum, pnum):
        confusions.draw_precision_recall_curve(fnum=fnum, pnum=pnum)

    if with_scores:
        inter.append_plot(SortedScoreSupportInteraction)
    if with_hist:
        inter.append_plot(_score_support_hist)
    if with_prebayes:
        inter.append_plot(_prebayes)
    if with_postbayes:
        inter.append_plot(_postbayes)
    if with_normscore:
        inter.append_plot(_prob_support_sorted)
        #inter.append_plot(_prob_support_hist)
    if with_roc:
        inter.append_plot(ROCInteraction)
    if with_precision_recall:
        inter.append_plot(_precision_recall)

    inter.start()

    if figtitle is not None:
        pt.set_figtitle(figtitle)

    return inter


def estimate_pdf(data, gridsize=1024, adjust=1):
    """
    estimate_pdf

    References;
        http://statsmodels.sourceforge.net/devel/generated/statsmodels.nonparametric.kde.KDEUnivariate.html
        https://jakevdp.github.io/blog/2013/12/01/kernel-density-estimation/

    Args:
        data (ndarray): 1 dimensional data of float64
        gridsize(int): domain size
        adjust(int): smoothing factor

    Returns:
        ndarray: data_pdf

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.score_normalization import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> rng = np.random.RandomState(0)
        >>> data = rng.randn(1000)
        >>> data_pdf = vt.estimate_pdf(data)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.plot(data_pdf.support[:-1], np.diff(data_pdf.cdf))
        >>> ut.show_if_requested()
    """
    import utool as ut
    import numpy as np
    import statsmodels.nonparametric.kde
    import statsmodels.nonparametric.bandwidths
    #import scipy.stats as spstats
    #import statsmodelskde.score_samples(support[:, None])

    if True:
        # Ensure that a non-zero bandwidth is chosen
        # bw_choices = ['scott', 'silverman', 'normal_reference']
        # bw = bw_choices[1]
        for bw in ['silverman', 'scott']:
            bw_value = statsmodels.nonparametric.bandwidths.select_bandwidth(data, bw, None)
            if bw_value > 0:
                break
        if bw_value == 0:
            sorted_diffs = np.diff(sorted(data))
            nonzero_diffs = sorted_diffs[sorted_diffs > 0]
            if len(nonzero_diffs) > 0:
                median_diff = np.median(nonzero_diffs)
                bw_value = np.sqrt(median_diff)
            else:
                # use a very small value
                bw_value = 1e-9
    try:

        if False:
            # Alternate implementation in case statsmodels breaks
            class TempPdf():
                def __init__(data_pdf, data, bw_value, gridsize):
                    from sklearn.neighbors.kde import KernelDensity
                    kde = KernelDensity(kernel='gaussian', bandwidth=bw_value)
                    kde.fit(data[:, None])
                    data_pdf.kde = kde
                    data_pdf.support = np.linspace(data.min(), data.max(), gridsize)
                    data_pdf.density = data_pdf.evaluate(data_pdf.support)
                    # import scipy as sp
                    data_pdf.cdf = data_pdf.density.cumsum()
                    # data_pdf.cdf = sp.integrate.cumtrapz(data_pdf.density,
                    #                                        data_pdf.support)

                def evaluate(data_pdf, scores):
                    return np.exp(data_pdf.kde.score_samples(scores[:, None]))
            data_pdf = TempPdf(data, bw_value, gridsize)

        data_pdf = statsmodels.nonparametric.kde.KDEUnivariate(data)
        fitkw = dict(kernel='gau',
                     bw=bw_value,
                     fft=True,
                     weights=None,
                     adjust=adjust,
                     cut=3,
                     gridsize=gridsize,
                     clip=(-np.inf, np.inf),)
        data_pdf.fit(**fitkw)
    except Exception as ex:
        ut.printex(ex, '! Exception while estimating kernel density',
                   keys=['data', 'gridsize', 'bw_value', 'adjust'])
        raise

    return data_pdf


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.score_normalization
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python
"""
wraps c implementations slower parts of spatial verification

CommandLine:
    python -m vtool_ibeis.sver_c_wrapper --rebuild-sver
    python -m vtool_ibeis.sver_c_wrapper --rebuild-sver --allexamples
    python -m vtool_ibeis.sver_c_wrapper --allexamples

    python -m vtool_ibeis.sver_c_wrapper --test-test_sver_wrapper --rebuild-sver

Example:
    >>> # xdoctest: +REQUIRES(module:vtool_ibeis_ext)
    >>> import numpy as np
    >>> kpts1 = np.array([[ 3.2e+01,  2.7e+01,  1.7e+01,  5.0e+00,  2.1e+01,  6.2e+00],
    >>>                   [ 1.3e+02,  2.3e+01,  2.0e+01,  2.7e+00,  2.1e+01,  6.3e+00],
    >>>                   [ 2.3e+02,  2.4e+01,  1.7e+01, -9.2e+00,  1.6e+01,  4.7e-01],
    >>>                   [ 3.0e+01,  1.3e+02,  1.6e+01, -9.2e+00,  1.8e+01,  5.6e+00],
    >>>                   [ 1.3e+02,  1.4e+02,  1.9e+01, -1.6e-01,  2.1e+01,  1.1e-01],
    >>>                   [ 2.3e+02,  1.3e+02,  1.9e+01, -1.1e+00,  2.0e+01,  4.8e-02],
    >>>                   [ 3.5e+01,  2.2e+02,  1.8e+01, -3.0e+00,  2.0e+01,  1.8e-01],
    >>>                   [ 1.4e+02,  2.3e+02,  1.8e+01, -2.4e+00,  2.0e+01,  6.0e+00],
    >>>                   [ 2.3e+02,  2.3e+02,  2.3e+01,  1.3e+00,  1.9e+01,  4.4e-01]], dtype=np.float64)
    >>> kpts2 = np.array([[3.4e+01, 2.8e+01, 2.0e+01, 1.6e+00, 1.7e+01, 1.8e-02],
    >>>                   [1.3e+02, 2.7e+01, 2.1e+01, 4.4e+00, 2.0e+01, 6.3e+00],
    >>>                   [2.3e+02, 2.6e+01, 2.0e+01, 3.8e+00, 2.0e+01, 7.3e-03],
    >>>                   [3.2e+01, 1.3e+02, 2.3e+01, 1.1e+00, 2.2e+01, 6.0e+00],
    >>>                   [1.2e+02, 1.3e+02, 1.9e+01, 1.6e+00, 2.3e+01, 6.2e+00],
    >>>                   [2.3e+02, 1.4e+02, 1.8e+01, 4.5e+00, 1.6e+01, 2.7e-01],
    >>>                   [3.3e+01, 2.3e+02, 2.0e+01, 2.0e+00, 1.9e+01, 1.7e-01],
    >>>                   [1.3e+02, 2.3e+02, 2.1e+01, 4.0e+00, 2.1e+01, 1.1e-01],
    >>>                   [2.3e+02, 2.3e+02, 1.6e+01, 2.5e+00, 2.2e+01, 6.2e+00]], dtype=np.float64)
    >>> fm = np.array([[0, 0],[1, 1],[2, 2],[3, 3],[4, 4],[5, 5],[6, 6],[7, 7],[8, 8]], dtype=np.int64)
    >>> fs = np.array([1., 1., 1., 1., 1., 1., 1., 1., 1.], dtype=np.float64)
    >>> xy_thresh_sqrd = 100
    >>> scale_thresh_sqrd = 100
    >>> ori_thresh = 100
    >>> from vtool_ibeis.sver_c_wrapper import get_affine_inliers_cpp
    >>> from vtool_ibeis.sver_c_wrapper import get_best_affine_inliers_cpp
    >>> out_inliers, out_errors, out_mats = get_affine_inliers_cpp(kpts1, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh_sqrd, ori_thresh)
    >>> out_inliers, out_errors, out_mats = get_best_affine_inliers_cpp(kpts1, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh_sqrd, ori_thresh)
"""
from __future__ import absolute_import, division, print_function


def get_affine_inliers_cpp(kpts1, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh_sqrd, ori_thresh):
    """
    Example:
        kpts1 =
    """
    from vtool_ibeis_ext import sver_c_wrapper
    out_inliers, out_errors, out_mats = sver_c_wrapper.get_affine_inliers_cpp(
        kpts2, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh_sqrd, ori_thresh)
    return out_inliers, out_errors, out_mats


def get_best_affine_inliers_cpp(kpts1, kpts2, fm, fs, xy_thresh_sqrd,
                                scale_thresh_sqrd, ori_thresh):
    from vtool_ibeis_ext import sver_c_wrapper
    out_inliers, out_errors, out_mat = sver_c_wrapper.get_best_affine_inliers_cpp(
        kpts1, kpts2, fm, fs, xy_thresh_sqrd, scale_thresh_sqrd, ori_thresh)
    return out_inliers, out_errors, out_mat


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.sver_c_wrapper
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
OLD MODULE, needs reimplemenetation of select features and deprication

This module should handle all things elliptical
"""
# from numpy.core.umath_tests import matrix_multiply
import operator as op
import scipy.signal as spsignal
import numpy as np
from vtool_ibeis import keypoint as ktool
from vtool_ibeis import image as gtool
try:
    import cv2
except ImportError:
    print('ERROR: import cv2 is failing!')
    cv2 = object()


def adaptive_scale(img_fpath, kpts, nScales=4, low=-.5, high=.5, nSamples=16):
    #imgBGR = cv2.imread(img_fpath, flags=cv2.CV_LOAD_IMAGE_COLOR)
    imgBGR = gtool.imread(img_fpath)

    nKp = len(kpts)
    dtype_ = kpts.dtype

    # Work with float65
    kpts_ = np.array(kpts, dtype=np.float64)

    # Expand each keypoint into a number of different scales
    expanded_kpts = expand_scales(kpts_, nScales, low, high)

    # Sample gradient magnitude around the border
    border_vals_sum = sample_ell_border_vals(imgBGR, expanded_kpts, nKp, nScales, nSamples)

    # interpolate maxima
    subscale_kpts = subscale_peaks(border_vals_sum, kpts_, nScales, low, high)

    # Make sure that the new shapes are in bounds
    height, width = imgBGR.shape[0:2]
    isvalid = check_kpts_in_bounds(subscale_kpts, width, height)

    # Convert to the original dtype
    adapted_kpts = np.array(subscale_kpts[isvalid], dtype=dtype_)
    return adapted_kpts


def check_kpts_in_bounds(kpts_, width, height):
    # Test to make sure the extents of the keypoints are in bounds
    unit_bbox = np.array([(-1, -1, 1),
                          (-1,  1, 1),
                          ( 1, -1, 1),
                          ( 1,  1, 1)]).T
    #invV = kpts_to_invV(kpts_)
    invV = ktool.get_invV_mats3x3(kpts_)
    bbox_pts = [v.dot(unit_bbox)[0:2] for v in invV]
    maxx = np.array([pts[0].max() for pts in bbox_pts]) < width
    minx = np.array([pts[0].min() for pts in bbox_pts]) > 0
    maxy = np.array([pts[1].max() for pts in bbox_pts]) < height
    miny = np.array([pts[1].min() for pts in bbox_pts]) > 0
    isvalid = np.array(maxx * minx * maxy * miny, dtype=bool)
    return isvalid


def expand_scales(kpts, nScales, low, high):
    scales = 2 ** np.linspace(low, high, nScales)
    expanded_kpts_list = expand_kpts(kpts, scales)
    expanded_kpts = np.vstack(expanded_kpts_list)
    #assert len(expanded_kpts_list) == nScales
    #assert expanded_kpts.shape == (nKp * nScales, 5)
    return expanded_kpts


def sample_ell_border_pts(expanded_kpts, nSamples):
    ell_border_pts_list = sample_uniform(expanded_kpts, nSamples)
    #assert len(ell_border_pts_list) == nKp * nScales
    #assert ell_border_pts_list[0].shape == (nSamples, 2)
    ell_border_pts = np.vstack(ell_border_pts_list)
    #assert ell_border_pts.shape == (nKp * nScales * nSamples, 2)
    #assert ell_border_pts.shape == (nKp * nScales * nSamples, 2)
    return ell_border_pts


def sample_ell_border_vals(imgBGR, expanded_kpts, nKp, nScales, nSamples):
    # Sample points uniformly across the boundary
    ell_border_pts = sample_ell_border_pts(expanded_kpts, nSamples)
    # Build gradient magnitude imaeg
    imgLAB = cv2.cvtColor(imgBGR, cv2.COLOR_BGR2LAB)
    imgL = imgLAB[:, :, 0]
    imgMag = gradient_magnitude(imgL)
    border_vals = gtool.subpixel_values(imgMag, ell_border_pts)
    #assert len(border_vals) == (nKp * nScales * nSamples)
    border_vals.shape = (nKp, nScales, nSamples, 1)
    border_vals_sum = border_vals.sum(3).sum(2)
    #assert border_vals_sum.shape == (nKp, nScales)
    return border_vals_sum


def interpolate_between(peak_list, nScales, high, low):
    def bin_to_subscale(peaks):
        return 2 ** ((peaks[:, 0] / nScales) * (high - low) + low)
    subscale_list = [bin_to_subscale(peaks) if len(peaks) > 0 else []
                     for peaks in peak_list]
    return subscale_list


def subscale_peaks(border_vals_sum, kpts, nScales, low, high):
    peak_list = interpolate_maxima(border_vals_sum)
    subscale_list = interpolate_between(peak_list, nScales, high, low)
    subscale_kpts = expand_subscales(kpts, subscale_list)
    return subscale_kpts


def expand_kpts(kpts, scales):
    expanded_kpts_list = []
    for scale in scales:
        kpts_ = kpts.copy()
        kpts_.T[2] *= scale
        kpts_.T[3] *= scale
        kpts_.T[4] *= scale
        expanded_kpts_list.append(kpts_)
    return expanded_kpts_list


def expand_subscales(kpts, subscale_list):
    subscale_kpts_list = [kp * np.array((1, 1, scale, scale, scale, 1))
                          for kp, subscales in zip(kpts, subscale_list)
                          for scale in subscales]
    subscale_kpts = np.vstack(subscale_kpts_list)
    return subscale_kpts


def find_maxima(y_list):
    maxima_list = [spsignal.argrelextrema(y, np.greater)[0] for y in y_list]
    return maxima_list


def extrema_neighbors(extrema_list, nBins):
    extrema_left_list  = [np.clip(extrema - 1, 0, nBins) for extrema in extrema_list]
    extrema_right_list = [np.clip(extrema + 1, 0, nBins) for extrema in extrema_list]
    return extrema_left_list, extrema_right_list


def find_maxima_with_neighbors(scalar_list):
    y_list = [scalars for scalars in scalar_list]
    nBins = len(y_list[0])
    x = np.arange(nBins)
    maxima_list = find_maxima(y_list)
    maxima_left_list, maxima_right_list = extrema_neighbors(maxima_list, nBins)
    data_list = [np.vstack([exl, exm, exr]) for exl, exm, exr in zip(maxima_left_list, maxima_list, maxima_right_list)]
    x_data_list = [[] if data.size == 0 else x[data] for data in iter(data_list)]
    y_data_list = [[] if data.size == 0 else y[data] for y, data in zip(y_list, data_list)]
    return x_data_list, y_data_list


def interpolate_maxima(scalar_list):
    # scalar_list = border_vals_sum
    x_data_list, y_data_list = find_maxima_with_neighbors(scalar_list)
    peak_list = interpolate_peaks(x_data_list, y_data_list)
    return peak_list


def interpolate_peaks2(x_data_list, y_data_list):
    coeff_list = []
    for x_data, y_data in zip(x_data_list, y_data_list):
        for x, y in zip(x_data.T, y_data.T):
            coeff = np.polyfit(x, y, 2)
            coeff_list.append(coeff)


def interpolate_peaks(x_data_list, y_data_list):
    #http://stackoverflow.com/questions/717762/how-to-calculate-the-vertex-of-a-parabola-given-three-point
    peak_list = []
    for x_data, y_data in zip(x_data_list, y_data_list):
        if len(y_data) == 0:
            peak_list.append([])
            continue
        y1, y2, y3 = y_data
        x1, x2, x3 = x_data
        denom = (x1 - x2) * (x1 - x3) * (x2 - x3)
        A     = (x3 * (y2 - y1) + x2 * (y1 - y3) + x1 * (y3 - y2)) / denom
        B     = (x3 * x3 * (y1 - y2) + x2 * x2 * (y3 - y1) + x1 * x1 * (y2 - y3)) / denom
        C     = (x2 * x3 * (x2 - x3) * y1 + x3 * x1 * (x3 - x1) * y2 + x1 * x2 * (x1 - x2) * y3) / denom
        xv = -B / (2 * A)
        yv = C - B * B / (4 * A)
        peak_list.append(np.vstack((xv.T, yv.T)).T)
    return peak_list


def sample_uniform(kpts, nSamples=128):
    """
    SeeAlso:
        python -m pyhesaff.tests.test_ellipse --test-in_depth_ellipse --show

    """
    nKp = len(kpts)
    # Get keypoint matrix forms
    invV_mats3x3 = ktool.get_invV_mats3x3(kpts)
    V_mats3x3 = ktool.invert_invV_mats(invV_mats3x3)
    #-------------------------------
    # Get uniform points on a circle
    circle_pts = homogenous_circle_pts(nSamples + 1)[0:-1]
    assert circle_pts.shape == (nSamples, 3)
    #-------------------------------
    # Get uneven points sample (get_uneven_point_sample)
    polygon1_list = op.matmul(invV_mats3x3, circle_pts.T).transpose(0, 2, 1)
    assert polygon1_list.shape == (nKp, nSamples, 3)
    # -------------------------------
    # The transformed points are not sampled uniformly... Bummer
    # We will sample points evenly across the sampled polygon
    # then we will project them onto the ellipse
    dists = np.array([circular_distance(arr) for arr in polygon1_list])
    assert dists.shape == (nKp, nSamples)
    # perimeter of the polygon
    perimeter = dists.sum(1)
    assert perimeter.shape == (nKp,)
    # Take a perfect multiple of steps along the perimeter
    multiplier = 1
    step_size = perimeter / (nSamples * multiplier)
    assert step_size.shape == (nKp,)
    # Walk along edge
    num_steps_list = []
    offset_list = []
    total_dist = np.zeros(step_size.shape)  # step_size.copy()
    dist_walked = np.zeros(step_size.shape)
    assert dist_walked.shape == (nKp,)
    assert total_dist.shape == (nKp,)
    distsT = dists.T
    assert distsT.shape == (nSamples, nKp)

    # This loops over the pt samples and performs the operation for every keypoint
    for count in range(nSamples):
        segment_len = distsT[count]
        # Find where your starting location is
        offset_list.append(total_dist - dist_walked)
        # How far can you possibly go?
        total_dist += segment_len
        # How many steps can you take?
        num_steps = (total_dist - dist_walked) // step_size
        num_steps_list.append(num_steps)
        # Log how much further youve gotten
        dist_walked += (num_steps * step_size)
    # Check for floating point errors
    # take an extra step if you need to
    num_steps_list[-1] += np.round((perimeter - dist_walked) / step_size)
    assert np.all(np.array(num_steps_list).sum(0) == nSamples)

    """
    #offset_iter1 = zip(num_steps_list, distsT, offset_list)
    #offset_list = [((step_size - offset) / dist, ((num * step_size) - offset) / dist, num)
                   #for num, dist, offset in zip(num_steps_list, distsT, offset_list)]

    #offset_iter2 = offset_list

    #cut_locs = [[
        #np.linspace(off1, off2, n, endpoint=True) for (off1, off2, n) in zip(offset1, offset2, num)]
        #for (offset1, offset2, num) in offset_iter2
    #]
    # store the percent location at each line segment where
    # the cut will be made
    """
    # HERE IS NEXT
    cut_list = []
    # This loops over the pt samples and performs the operation for every keypoint
    for num, dist, offset in zip(num_steps_list, distsT, offset_list):
        #if num == 0
        #  cut_list.append([])
        #  continue
        # This was a bitch to keep track of
        offset1 = (step_size - offset) / dist
        offset2 = ((num * step_size) - offset) / dist
        cut_locs = [np.linspace(off1, off2, int(n), endpoint=True) for (off1, off2, n) in zip(offset1, offset2, num)]
        # post check for divide by 0
        cut_locs = [np.array([0 if np.isinf(c) else c for c in cut]) for cut in cut_locs]
        cut_list.append(cut_locs)
    cut_list = np.array(cut_list).T
    assert cut_list.shape == (nKp, nSamples)

    # =================
    # METHOD 1
    # =================
    # Linearly interpolate between points on the polygons at the places we cut
    def interpolate(pt1, pt2, percent):
        # interpolate between point1 and point2
        return ((1 - percent) * pt1) + ((percent) * pt2)

    def polygon_points(polygon_pts, dist_list):
        return np.array([interpolate(polygon_pts[count], polygon_pts[(count + 1) % nSamples], loc)
                         for count, locs in enumerate(dist_list)
                         for loc in iter(locs)])

    new_locations = np.array([polygon_points(polygon_pts, cuts) for polygon_pts,
                              cuts in zip(polygon1_list, cut_list)])
    # =================
    # =================
    # METHOD 2
    """
    #from itertools import cycle as icycle
    #from itertools import islice

    #def icycle_shift1(iterable):
        #return islice(icycle(poly_pts), 1, len(poly_pts) + 1)

    #cutptsIter_list = [zip(iter(poly_pts), icycle_shift1(poly_pts), cuts)
                       #for poly_pts, cuts in zip(polygon1_list, cut_list)]
    #new_locations = [[[((1 - cut) * pt1) + ((cut) * pt2) for cut in cuts]
                      #for (pt1, pt2, cuts) in cutPtsIter]
                     #for cutPtsIter in cutptsIter_list]
    """
    # =================

    # assert new_locations.shape == (nKp, nSamples, 3)
    # Warp new_locations to the unit circle
    #new_unit = V.dot(new_locations.T).T
    new_unit = np.array([v.dot(newloc.T).T for v, newloc in zip(V_mats3x3, new_locations)])
    # normalize new_unit
    new_mag = np.sqrt((new_unit ** 2).sum(-1))
    new_unorm_unit = new_unit / np.dstack([new_mag] * 3)
    new_norm_unit = new_unorm_unit / np.dstack([new_unorm_unit[:, :, 2]] * 3)
    # Get angle (might not be necessary)
    #x_axis = np.array([1, 0, 0])
    #arccos_list = x_axis.dot(new_norm_unit.T)
    #uniform_theta_list = np.arccos(arccos_list)
    # Maybe this?
    # Find the angle from the center of the circle
    theta_list2 = np.arctan2(new_norm_unit[:, :, 1], new_norm_unit[:, :, 0])
    # assert uniform_theta_list.shape = (nKp, nSample)
    # Use this angle to unevenly sample the perimeter of the circle
    uneven_cicrle_pts = np.dstack([np.cos(theta_list2), np.sin(theta_list2), np.ones(theta_list2.shape)])
    # The uneven circle points were sampled in such a way that when they are
    # transformeed they will be approximately uniform along the boundary of the
    # ellipse.
    uniform_ell_hpts = [v.dot(pts.T).T for (v, pts) in zip(invV_mats3x3, uneven_cicrle_pts)]
    # Remove the homogenous coordinate and we're done
    ell_border_pts_list = [pts[:, 0:2] for pts in uniform_ell_hpts]
    return ell_border_pts_list


#----------------
# Image Helpers
#----------------


def gradient_magnitude(img):
    sobelx = cv2.Sobel(img, cv2.CV_64F, 1, 0, ksize=3)
    sobely = cv2.Sobel(img, cv2.CV_64F, 0, 1, ksize=3)
    imgMag = np.sqrt(sobelx ** 2 + sobely ** 2)
    return imgMag


#----------------
# Numeric Helpers
#----------------

#def kpts_to_invV(kpts):
#    invV = ktool.get_invV_mats3x3(kpts)
#    #invV = ktool.get_invV_mats(kpts, ashomog=True,
#    #                                with_trans=True, ascontiguous=True)
#    return invV
#    #nKp = len(kpts)
#    #(iv13, iv23, iv11, iv21, iv22) = np.array(kpts).T
#    #iv12 = zeros(nKp)
#    #iv31 = zeros(nKp)
#    #iv32 = zeros(nKp)
#    #iv33 = ones(nKp)
#    ## np.dot operates over the -1 and -2 axis of arrays:27
#    ## Start with
#    ##     invV.shape = (3, 3, nKp)
#    #invV = np.array([[iv11, iv12, iv13],
#    #              [iv21, iv22, iv23],
#    #              [iv31, iv32, iv33]])
#    ## And roll into
#    #invV = np.rollaxis(invV, 2)
#    #invV = np.ascontiguousarray(invV)
#    #assert invV.shape == (nKp, 3, 3)
#    #return invV


def kpts_matrices(kpts):
    # We are given the keypoint in invA format
    # invV = perdoch.invA
    #    V = perdoch.A
    #    Z = perdoch.E
    # invert into V
    #invV = kpts_to_invV(kpts)
    invV = ktool.get_invV_mats3x3(kpts)
    V = ktool.invert_invV_mats(invV)
    Z = ktool.get_Z_mats(V)
    return invV, V, Z


def homogenous_circle_pts(nSamples):
    """  Make a list of homogenous circle points """
    tau = 2 * np.pi
    theta_list = np.linspace(0, tau, nSamples)
    circle_pts = np.array([(np.cos(t_), np.sin(t_), 1) for t_ in theta_list])
    return circle_pts


def circular_distance(arr=None):
    dist_head_ = ((arr[0:-1] - arr[1:]) ** 2).sum(1).astype(float)
    dist_tail_  = ((arr[-1] - arr[0]) ** 2).sum(0).astype(float)
    #dist_end_.shape = (1, len(dist_end_))
    #print(dist_most_.shape)
    #print(dist_end_.shape)
    dists = np.sqrt(np.hstack((dist_head_, dist_tail_)))
    return dists



import warnings
import scipy.signal
import numpy as np
import utool as ut
from .util_math import TAU


def argsubmax(ydata, xdata=None):
    """
    Finds a single submaximum value to subindex accuracy.
    If xdata is not specified, submax_x is a fractional index.
    Otherwise, submax_x is sub-xdata (essentially doing the index interpolation
    for you)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> import ubelt as ub
        >>> ydata = [ 0,  1,  2, 1.5,  0]
        >>> xdata = [00, 10, 20,  30, 40]
        >>> result1 = argsubmax(ydata, xdata=None)
        >>> result2 = argsubmax(ydata, xdata=xdata)
        >>> result = ub.repr2([result1, result2], precision=4, nl=1, nobr=True)
        >>> print(result)
        2.1667, 2.0208,
        21.6667, 2.0208,

    Example:
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> hist_ = np.array([0, 1, 2, 3, 4])
        >>> centers = None
        >>> maxima_thresh=None
        >>> argsubmax(hist_)
        (4.0, 4.0)
    """
    if len(ydata) == 0:
        raise IndexError('zero length array')
    ydata = np.asarray(ydata)
    xdata = None if xdata is None else np.asarray(xdata)
    submaxima_x, submaxima_y = argsubmaxima(ydata, centers=xdata)
    idx = submaxima_y.argmax()
    submax_y = submaxima_y[idx]
    submax_x = submaxima_x[idx]
    return submax_x, submax_y


def argsubmaxima(hist, centers=None, maxima_thresh=None, _debug=False):
    r"""
    Determines approximate maxima values to subindex accuracy.

    Args:
        hist_ (ndarray): ydata, histogram frequencies
        centers (ndarray): xdata, histogram labels
        maxima_thresh (float): cutoff point for labeing a value as a maxima

    Returns:
        tuple: (submaxima_x, submaxima_y)

    CommandLine:
        python -m vtool_ibeis.histogram argsubmaxima
        python -m vtool_ibeis.histogram argsubmaxima --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> maxima_thresh = .8
        >>> hist = np.array([6.73, 8.69, 0.00, 0.00, 34.62, 29.16, 0.00, 0.00, 6.73, 8.69])
        >>> centers = np.array([-0.39, 0.39, 1.18, 1.96,  2.75,  3.53, 4.32, 5.11, 5.89, 6.68])
        >>> (submaxima_x, submaxima_y) = argsubmaxima(hist, centers, maxima_thresh)
        >>> result = str((submaxima_x, submaxima_y))
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.draw_hist_subbin_maxima(hist, centers)
        >>> pt.show_if_requested()
        (array([ 3.0318792]), array([ 37.19208239]))
    """
    maxima_x, maxima_y, argmaxima = hist_argmaxima(hist, centers, maxima_thresh=maxima_thresh)
    argmaxima = np.asarray(argmaxima)
    if _debug:
        print('Argmaxima: ')
        print(' * maxima_x = %r' % (maxima_x))
        print(' * maxima_y = %r' % (maxima_y))
        print(' * argmaxima = %r' % (argmaxima))
    flags = (argmaxima == 0) | (argmaxima == len(hist) - 1)
    argmaxima_ = argmaxima[~flags]
    submaxima_x_, submaxima_y_ = interpolate_submaxima(argmaxima_, hist, centers)
    if np.any(flags):
        endpts = argmaxima[flags]
        submaxima_x = (np.hstack([submaxima_x_, centers[endpts]])
                       if centers is not None else
                       np.hstack([submaxima_x_, endpts]))
        submaxima_y = np.hstack([submaxima_y_, hist[endpts]])
    else:
        submaxima_y = submaxima_y_
        submaxima_x = submaxima_x_
    if _debug:
        print('Submaxima: ')
        print(' * submaxima_x = %r' % (submaxima_x))
        print(' * submaxima_y = %r' % (submaxima_y))
    return submaxima_x, submaxima_y


def argsubmin2(ydata, xdata=None):
    if len(ydata) == 0:
        raise ValueError('zero length array')
    ydata = np.asarray(ydata)
    xdata = None if xdata is None else np.asarray(xdata)
    submax_x, submax_y = argsubmax2(-ydata, xdata)
    submin_x = submax_x
    submin_y = -submax_y
    return submin_x, submin_y


def argsubmax2(ydata, xdata=None):
    """
    Finds a single submaximum value to subindex accuracy.
    If xdata is not specified, submax_x is a fractional index.
    This version always normalizes x-coordinates.

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> ydata = [ 0,  1,  2, 1.5,  0]
        >>> xdata = [00, 10, 20,  30, 40]
        >>> result1 = argsubmax(ydata, xdata=None)
        >>> result2 = argsubmax(ydata, xdata=xdata)
        >>> import ubelt as ub
        >>> result = ub.repr2([result1, result2], precision=4, nl=1, nobr=True)
        >>> print(result)

        (2.1667, 2.0208),
        (21.6667, 2.0208),

    Example:
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> hist_ = np.array([0, 1, 2, 3, 4])
        >>> centers = None
        >>> thresh_factor = None
        >>> argsubmax(hist_)

        (4.0, 4.0)
    """
    if len(ydata) == 0:
        raise ValueError('zero length array')
    ydata = np.asarray(ydata)
    xdata = None if xdata is None else np.asarray(xdata)
    submaxima_x, submaxima_y = argsubmaxima2(ydata, xdata=xdata,
                                             normalize_x=True)
    idx = submaxima_y.argmax()
    submax_y = submaxima_y[idx]
    submax_x = submaxima_x[idx]
    return submax_x, submax_y


def argsubmaxima2(ydata, xdata=None, thresh_factor=None, normalize_x=True):
    return argsubextrema2('max', ydata, xdata=xdata,
                          thresh_factor=thresh_factor, normalize_x=normalize_x)


def argsubminima2(ydata, xdata=None, thresh_factor=None, normalize_x=True):
    """
    """
    return argsubextrema2('min', ydata, xdata=xdata,
                          thresh_factor=thresh_factor, normalize_x=normalize_x)


def argsubextrema2(op, ydata, xdata=None, thresh_factor=None, normalize_x=True,
                   flat=True):
    r"""
    Determines approximate maxima values to subindex accuracy.

    Args:
        ydata (ndarray): ydata, histogram frequencies
        xdata (ndarray): xdata, histogram labels
        thresh_factor (float): cutoff point for labeing a value as a maxima
        flat (bool): if True allows for flat extrema to be found.

    Returns:
        tuple: (submaxima_x, submaxima_y)

    CommandLine:
        python -m vtool_ibeis.histogram argsubmaxima
        python -m vtool_ibeis.histogram argsubmaxima --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> thresh_factor = .8
        >>> ydata = np.array([6.73, 8.69, 0.00, 0.00, 34.62, 29.16, 0.00, 0.00, 6.73, 8.69])
        >>> xdata = np.array([-0.39, 0.39, 1.18, 1.96,  2.75,  3.53, 4.32, 5.11, 5.89, 6.68])
        >>> op = 'min'
        >>> (subextrema_x, subextrema_y) = argsubextrema2(op, ydata, xdata, thresh_factor)
        >>> result = str((subextrema_x, subextrema_y))
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.draw_hist_subbin_maxima(ydata, xdata)
        >>> pt.show_if_requested()

    Doctest:
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> thresh_factor = .8
        >>> ydata = np.array([1, 1, 1, 2, 1, 2, 3, 2, 4, 1.1, 5, 1.2, 1.1, 1.1, 1.2, 1.1])
        >>> op = 'max'
        >>> thresh_factor = .8
        >>> (subextrema_x, subextrema_y) = argsubextrema2(op, ydata, thresh_factor=thresh_factor)
        >>> result = str((subextrema_x, subextrema_y))
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.qtensure()
        >>> xdata = np.arange(len(ydata))
        >>> pt.figure(fnum=1, doclf=True)
        >>> pt.plot(xdata, ydata)
        >>> pt.plot(subextrema_x, subextrema_y, 'o')
        >>> ut.show_if_requested()
    """
    if op == 'max':
        cmp = np.greater
        cmp_eq = np.greater_equal
        extreme = np.max
        factor_op = np.multiply
    elif op == 'min':
        cmp = np.less
        cmp_eq = np.less_equal
        extreme = np.min
        factor_op = np.divide
    else:
        raise KeyError(op)

    # absolute extrema
    abs_extreme_y = extreme(ydata)

    if thresh_factor is None:
        thresh_factor = 1.0
        # thresh_factor = None

    if thresh_factor is None:
        thresh_value = None
        flags = np.ones(len(ydata), dtype=bool)
    else:
        # Find relative and flat extrema
        thresh_value = factor_op(abs_extreme_y, thresh_factor)
        flags = cmp_eq(ydata, thresh_value)

    # Only consider non-boundary points
    flags[[0, -1]] = False
    mxs = np.where(flags)[0]
    # ydata[np.vstack([mxs - 1, mxs, mxs + 1])]
    lvals, mvals, rvals = ydata[mxs - 1], ydata[mxs], ydata[mxs + 1]
    is_mid_extrema = cmp_eq(mvals, lvals) & cmp_eq(mvals, rvals)
    is_rel_extrema = cmp(mvals, lvals) & cmp(mvals, rvals)
    is_flat_extrema = is_mid_extrema & ~is_rel_extrema

    flat_argextrema = mxs[is_flat_extrema]
    rel_argextrema = mxs[is_rel_extrema]

    # Sublocalize relative extrema
    if len(rel_argextrema) > 0:
        # rel_subextrema_x, rel_subextrema_y = _sublocalize_extrema(
        #     ydata, xdata, rel_argextrema, cmp, normalize_x)
        neighbs = np.vstack((rel_argextrema - 1, rel_argextrema, rel_argextrema + 1))
        y123 = ydata[neighbs]
        if normalize_x or xdata is None:
            x123 = neighbs
        else:
            x123 = xdata[neighbs]

        # Fit parabola around points
        coeff_list = []
        for (x, y) in zip(x123.T, y123.T):
            coeff = np.polyfit(x, y, deg=2)
            coeff_list.append(coeff)

        A, B, C = np.vstack(coeff_list).T

        # Extreme x point is where the derivative is 0
        rel_subextrema_x = -B / (2 * A)
        rel_subextrema_y = C - B * B / (4 * A)

        if xdata is not None and normalize_x:
            # Convert x back to data coordinates if we normalized durring polynimal
            # fitting. Do linear interpoloation.
            rel_subextrema_x = linear_interpolation(xdata, rel_subextrema_x)

        # Check to make sure rel_subextrema is not less extreme than the original
        # extrema (can be the case only if the extrema is incorrectly given)
        # In this case just return what the user wanted as the extrema
        violates = cmp(ydata[rel_argextrema], rel_subextrema_y)
        if np.any(violates):
            warnings.warn(
                'rel_subextrema was less extreme than the measured extrema',
                RuntimeWarning)
            if xdata is not None:
                rel_subextrema_x[violates] = xdata[rel_argextrema[violates]]
            else:
                rel_subextrema_x[violates] = rel_argextrema[violates]
            rel_subextrema_y[violates] = ydata[rel_argextrema[violates]]
    else:
        x123 = None  # for plottool_ibeis
        coeff_list = []  # for plottool_ibeis
        rel_subextrema_x, rel_subextrema_y = [], []

    # Check left boundary
    boundary_argextrema = []
    if len(ydata) > 1:
        if thresh_value is None or cmp_eq(ydata[0], thresh_value):
            if cmp_eq(ydata[0], ydata[1]):
                boundary_argextrema.append(0)

    # Check right boundary
    if len(ydata) > 2:
        if thresh_value is None or cmp_eq(ydata[0], thresh_value):
            if cmp_eq(ydata[-1], ydata[-2]):
                boundary_argextrema.append(len(ydata) - 1)

    # Any non-flat mid extrema can be sublocalized
    other_argextrema = np.hstack([boundary_argextrema, flat_argextrema])
    other_argextrema = other_argextrema.astype(int)
    other_subextrema_y = ydata[other_argextrema]
    if xdata is None:
        other_subextrema_x = other_argextrema
    else:
        other_subextrema_x = xdata[other_argextrema]

    # Recombine and order all extremas
    subextrema_y = np.hstack([rel_subextrema_y, other_subextrema_y])
    subextrema_x = np.hstack([rel_subextrema_x, other_subextrema_x])
    sortx = subextrema_x.argsort()
    subextrema_x = subextrema_x[sortx]
    subextrema_y = subextrema_y[sortx]
    return subextrema_x, subextrema_y


def _sublocalize_extrema(ydata, xdata, argextrema, cmp, normalize_x=True):
    """
    Assumes argextrema are all relative and not on the boundaries
    """
    # extrema_y = ydata[argextrema]
    # extrema_x = argextrema if xdata is None else xdata[argextrema]

    neighbs = np.vstack((argextrema - 1, argextrema, argextrema + 1))
    y123 = ydata[neighbs]
    if normalize_x or xdata is None:
        x123 = neighbs
    else:
        x123 = xdata[neighbs]

    # Fit parabola around points
    coeff_list = []
    for (x, y) in zip(x123.T, y123.T):
        coeff = np.polyfit(x, y, deg=2)
        coeff_list.append(coeff)

    A, B, C = np.vstack(coeff_list).T

    # Extreme x point is where the derivative is 0
    subextrema_x = -B / (2 * A)
    subextrema_y = C - B * B / (4 * A)

    if xdata is not None and normalize_x:
        # Convert x back to data coordinates if we normalized durring polynimal
        # fitting. Do linear interpoloation.
        subextrema_x = linear_interpolation(xdata, subextrema_x)

    # Check to make sure subextrema is not less extreme than the original
    # extrema (can be the case only if the extrema is incorrectly given)
    # In this case just return what the user wanted as the extrema
    violates = cmp(ydata[argextrema], subextrema_y)
    if np.any(violates):
        warnings.warn(
            'subextrema was less extreme than the measured extrema',
            RuntimeWarning)
        if xdata is not None:
            subextrema_x[violates] = xdata[argextrema[violates]]
        else:
            subextrema_x[violates] = argextrema[violates]
        subextrema_y[violates] = ydata[argextrema[violates]]
    return subextrema_x, subextrema_y


def linear_interpolation(arr, subindices):
    """
    Does linear interpolation to lookup subindex values

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> arr = np.array([0, 1, 2, 3])
        >>> subindices = np.array([0, .1, 1, 1.8, 2, 2.5, 3] )
        >>> subvalues = linear_interpolation(arr, subindices)
        >>> assert np.allclose(subindices, subvalues)
        >>> assert np.allclose(2.3, linear_interpolation(arr, 2.3))
    """
    idx1 = np.floor(subindices).astype(int)
    idx2 = np.floor(subindices + 1).astype(int)
    idx2 = np.minimum(idx2, len(arr) - 1)
    alpha = idx2 - subindices
    subvalues = arr[idx1] * (alpha) + arr[idx2] * (1 - alpha)
    return subvalues


# subindex_take = linear_interpolation


def hist_argmaxima(hist, centers=None, maxima_thresh=None):
    """
    must take positive only values

    CommandLine:
        python -m vtool_ibeis.histogram hist_argmaxima

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> maxima_thresh = .8
        >>> hist = np.array([    6.73, 8.69, 0.00, 0.00, 34.62, 29.16, 0.00, 0.00, 6.73, 8.69])
        >>> centers = np.array([-0.39, 0.39, 1.18, 1.96,  2.75,  3.53, 4.32, 5.11, 5.89, 6.68])
        >>> maxima_x, maxima_y, argmaxima = hist_argmaxima(hist, centers)
        >>> result = str((maxima_x, maxima_y, argmaxima))
        >>> print(result)
    """
    # FIXME: Not handling general cases
    # [0] index because argrelmaxima returns a tuple
    argmaxima_ = scipy.signal.argrelextrema(hist, np.greater)[0]
    if len(argmaxima_) == 0:
        argmaxima_ = hist.argmax()
    if maxima_thresh is not None:
        # threshold maxima to be within a factor of the maximum
        maxima_y = hist[argmaxima_]
        isvalid = maxima_y > maxima_y.max() * maxima_thresh
        argmaxima = argmaxima_[isvalid]
    else:
        argmaxima = argmaxima_
    maxima_y = hist[argmaxima]
    maxima_x = argmaxima if centers is None else centers[argmaxima]
    return maxima_x, maxima_y, argmaxima


def hist_argmaxima2(hist, maxima_thresh=.8):
    """
    must take positive only values

    Setup:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA

    GridSearch:
        >>> hist1 = np.array([1, .9, .8, .99, .99, 1.1, .9, 1.0, 1.0])
        >>> hist2 = np.array([1, .9, .8, .99, .99, 1.1, 1.0, 1.0])
        >>> hist2 = np.array([1, .9, .8, .99, .99, 1.1, 1.0])
        >>> hist2 = np.array([1, .9, .8, .99, .99, 1.1, 1.2])
        >>> hist2 = np.array([1, 1.2])
        >>> hist2 = np.array([1, 1, 1.2])
        >>> hist2 = np.array([1])
        >>> hist2 = np.array([])

    Example:
        >>> # ENABLE_DOCTEST
        >>> maxima_thresh = .8
        >>> hist = np.array([1, .9, .8, .99, .99, 1.1, .9, 1.0, 1.0])
        >>> argmaxima = hist_argmaxima2(hist)
        >>> print(argmaxima)
    """
    # FIXME: Not handling general cases
    # [0] index because argrelmaxima returns a tuple
    if len(hist) == 0:
        return np.empty(dtype=int)
    comperetor = np.greater
    argmaxima_ = scipy.signal.argrelextrema(hist, comperetor)[0]
    if len(argmaxima_) == 0:
        argmaxima_ = np.array([hist.argmax()])  # Hack for no maxima
    maxval = hist[argmaxima_].max()
    size = len(hist)
    end = size - 1
    # Test if 0 is a maximum point
    if 0 not in argmaxima_ and size > 0:
        start_is_extreme = hist[0] > hist[1]
        if start_is_extreme and hist[0] >= maxval * maxima_thresh:
            argmaxima_ = np.hstack([[0], argmaxima_])
    # Test if end is maximum point
    if end not in argmaxima_ and end > 0:
        #end_is_extreme = np.all(hist[argmaxima_[-1] + 1:(end - 1)] < hist[end] )
        end_is_extreme = hist[end] > hist[end - 1]
        if not end_is_extreme:
            # FIXME: might be a case when end is level
            pass
            #end_is_extreme = np.all(hist[argmaxima_[-1] + 1:(end - 1)] == hist[end] )
        if end_is_extreme and hist[end] >= maxval * maxima_thresh:
            argmaxima_ = np.hstack([argmaxima_, [end]])
    # threshold maxima to be within a factor of the maximum
    maxima_y = hist[argmaxima_]
    isvalid = maxima_y >= maxval * maxima_thresh
    argmaxima = argmaxima_[isvalid]
    return argmaxima


def interpolate_submaxima(argmaxima, hist_, centers=None):
    r"""
    Args:
        argmaxima (ndarray): indicies into ydata / centers that are argmaxima
        hist_ (ndarray): ydata, histogram frequencies
        centers (ndarray): xdata, histogram labels

    FIXME:
        what happens when argmaxima[i] == len(hist_)

    CommandLine:
        python -m vtool_ibeis.histogram --test-interpolate_submaxima --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> argmaxima = np.array([1, 4, 7])
        >>> hist_ = np.array([    6.73, 8.69, 0.00, 0.00, 34.62, 29.16, 0.00, 0.00, 6.73, 8.69])
        >>> centers = np.array([-0.39, 0.39, 1.18, 1.96,  2.75,  3.53, 4.32, 5.11, 5.89, 6.68])
        >>> submaxima_x, submaxima_y = interpolate_submaxima(argmaxima, hist_, centers)
        >>> locals_ = ut.exec_func_src(interpolate_submaxima,
        >>>                            key_list=['x123', 'y123', 'coeff_list'])
        >>> x123, y123, coeff_list = locals_
        >>> res = (submaxima_x, submaxima_y)
        >>> import ubelt as ub
        >>> result = ub.repr2(res, nl=1, nobr=True, precision=2, with_dtype=True)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.ensureqt()
        >>> pt.figure(fnum=pt.ensure_fnum(None))
        >>> pt.plot(centers, hist_, '-')
        >>> pt.plot(centers[argmaxima], hist_[argmaxima], 'o', label='argmaxima')
        >>> pt.plot(submaxima_x, submaxima_y, 'b*', markersize=20, label='interp maxima')
        >>> # Extract parabola points
        >>> pt.plt.plot(x123, y123, 'o', label='maxima neighbors')
        >>> xpoints = [np.linspace(x1, x3, 50) for (x1, x2, x3) in x123.T]
        >>> ypoints = [np.polyval(coeff, x_pts) for x_pts, coeff in zip(xpoints, coeff_list)]
        >>> # Draw Submax Parabola
        >>> for x_pts, y_pts in zip(xpoints, ypoints):
        >>>     pt.plt.plot(x_pts, y_pts, 'g--', lw=2)
        >>> pt.show_if_requested()
        np.array([ 0.15,  3.03,  5.11], dtype=np.float64),
        np.array([  9.2 ,  37.19,   0.  ], dtype=np.float64),

    Example:
        >>> hist_ = np.array([5])
        >>> argmaxima = [0]

    """
    if len(argmaxima) == 0:
        return [], []
    argmaxima = np.asarray(argmaxima)
    neighbs = np.vstack((argmaxima - 1, argmaxima, argmaxima + 1))
    # flags = (neighbs[2] > (len(hist_) - 1)) | (neighbs[0] < 0)
    # neighbs = np.clip(neighbs, 0, len(hist_) - 1)
    # if np.any(flags):
    #     # Clip out of bounds positions
    #     neighbs[0, flags] = neighbs[1, flags]
    #     neighbs[2, flags] = neighbs[1, flags]
    y123 = hist_[neighbs]
    x123 = neighbs if centers is None else centers[neighbs]
    # if np.any(flags):
    #     # Make symetric values so maxima is found exactly in center
    #     y123[0, flags] = y123[1, flags] - 1
    #     y123[2, flags] = y123[1, flags] - 1
    #     x123[0, flags] = x123[1, flags] - 1
    #     x123[2, flags] = x123[1, flags] - 1
    # Fit parabola around points
    coeff_list = [np.polyfit(x123_, y123_, deg=2)
                  for (x123_, y123_) in zip(x123.T, y123.T)]
    A, B, C = np.vstack(coeff_list).T
    submaxima_x, submaxima_y = maximum_parabola_point(A, B, C)

    # Check to make sure submaxima is not less than original maxima
    # (can be the case only if the maxima is incorrectly given)
    # In this case just return what the user wanted as the maxima
    maxima_y = y123[1, :]
    invalid = submaxima_y < maxima_y
    if np.any(invalid):
        if centers is not None:
            submaxima_x[invalid] = centers[argmaxima[invalid]]
        else:
            submaxima_x[invalid] = argmaxima[invalid]
        submaxima_y[invalid] = hist_[argmaxima[invalid]]
    return submaxima_x, submaxima_y


def show_hist_submaxima(hist_, edges=None, centers=None, maxima_thresh=.8, pnum=(1, 1, 1)):
    r"""
    For C++ to show data

    Args:
        hist_ (?):
        edges (None):
        centers (None):

    CommandLine:
        python -m vtool_ibeis.histogram --test-show_hist_submaxima --show
        python -m pyhesaff._pyhesaff --test-test_rot_invar --show
        python -m vtool_ibeis.histogram --test-show_hist_submaxima --dpath figures --save ~/latex/crall-candidacy-2015/figures/show_hist_submaxima.jpg

    Example:
        >>> # DISABLE_DOCTEST
        >>> import plottool_ibeis as pt
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> hist_ = np.array(list(map(float, ut.get_argval('--hist', type_=list, default=[1, 4, 2, 5, 3, 3]))))
        >>> edges = np.array(list(map(float, ut.get_argval('--edges', type_=list, default=[0, 1, 2, 3, 4, 5, 6]))))
        >>> maxima_thresh = ut.get_argval('--maxima_thresh', type_=float, default=.8)
        >>> centers = None
        >>> show_hist_submaxima(hist_, edges, centers, maxima_thresh)
        >>> pt.show_if_requested()
    """
    import plottool_ibeis as pt
    if centers is None:
        centers = hist_edges_to_centers(edges)
    bin_colors = pt.get_orientation_color(centers)
    pt.figure(fnum=pt.next_fnum(), pnum=pnum)
    POLAR = False
    if POLAR:
        pt.df2.plt.subplot(*pnum, polar=True, axisbg='#000000')
    pt.draw_hist_subbin_maxima(hist_, centers, bin_colors=bin_colors, maxima_thresh=maxima_thresh)
    #pt.gca().set_rmax(hist_.max() * 1.1)
    #pt.gca().invert_yaxis()
    #pt.gca().invert_xaxis()
    pt.dark_background()
    #if ut.get_argflag('--legend'):
    #    pt.figure(fnum=pt.next_fnum())
    #    centers_ = np.append(centers, centers[0])
    #    r = np.ones(centers_.shape) * .2
    #    ax = pt.df2.plt.subplot(111, polar=True)
    #    pt.plots.colorline(centers_, r, cmap=pt.df2.plt.get_cmap('hsv'), linewidth=10)
    #    #ax.plot(centers_, r, 'm', color=bin_colors, linewidth=100)
    #    ax.set_rmax(.2)
    #    #ax.grid(True)
    #    #ax.set_title("Angle Colors", va='bottom')
    title = ut.get_argval('--title', default='')
    import plottool_ibeis as pt
    pt.set_figtitle(title)


def get_histinfo_str(hist, edges):
    centers = hist_edges_to_centers(edges)
    hist_str   = 'hist    = ' + str(hist.tolist())
    center_str = 'centers = ' + str(centers.tolist())
    edge_str   = 'edges   = [' +  ', '.join(['%.2f' % _ for _ in edges]) + ']'
    histinfo_str = hist_str + ut.NEWLINE + center_str + ut.NEWLINE + edge_str
    return histinfo_str


# def interpolated_wrapped_histogram2():
#     # another stab at it
#     data = np.array([ 0, .5,  2,  3.5,  3,  3,  3.9,  4])
#     weights = np.ones(len(data))

#     n_bins = 4
#     # max_val = np.pi * 2
#     max_val = 4
#     # Define the edges of each bin (we will vote into the middle)
#     # The first and last item are equivalent

#     # bin_centers = np.linspace(0, max_val, n_bins + 1, endpoint=True)
#     bin_centers = np.array([0, 1.5, 2, 3.5, 4])
#     bin_offsets = np.diff(bin_centers) / 2

#     # sort input data
#     sortx = np.argsort(data)
#     sa = data[sortx]


#     right_edges = bin_centers[:-1] + bin_offsets
#     wrapped_right = bin_centers[-1] + bin_offsets[0]

#     # Find data in the wrap-around zone
#     is_right_wrapped = (sa > right_edges[-1])

#     rectified_left = sa[is_right_wrapped] - max_val
#     # rectified_right = max_val - data[is_left_wrapped]

#     right_idx = right_edges.searchsorted(sa)



#     right_edges[-1] - data[]

#     data % right_edges[-1]

#     left_offsets = np.hstack([bin_offsets[-1], bin_offsets])
#     right_offsets = np.hstack([bin_offsets, bin_offsets[0]])

#     left_edges = bin_centers - left_offsets
#     right_edges = bin_centers + right_offsets


#     bin_edges = np.roll(bin_centers[1:] + bin_offsets, 1)


#     first_left_edge =

#     rssign_left = left_edges.searchsorted(sa)
#     assign_right = right_edges.searchsorted(sa)

#     # Find bins to the left and right of each point
#     assign_left  = sa.searchsorted(bin_edges[:-1], 'left')
#     assign_right = sa.searchsorted(bin_edges[-1], 'right')


def interpolated_histogram(data, weights, range_, bins,
                           interpolation_wrap=True, _debug=False):
    r"""
    Follows np.histogram, but does interpolation

    Args:
        data (ndarray):
        weights (ndarray):
        range_ (tuple): range from 0 to 1
        bins (int):
        interpolation_wrap (bool): (default = True)
        _debug (bool): (default = False)

    CommandLine:
        python -m vtool_ibeis.histogram --test-interpolated_histogram

    Example0:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> data = np.array([ 0,  1,  2,  3.5,  3,  3,  4,  4])
        >>> weights = np.array([1., 1., 1., 1., 1., 1., 1., 1.])
        >>> range_ = (0, 4)
        >>> bins = 5
        >>> interpolation_wrap = False
        >>> hist, edges = interpolated_histogram(data, weights, range_, bins,
        >>>                                      interpolation_wrap)
        >>> assert np.abs(hist.sum() - weights.sum()) < 1E-9
        >>> assert hist.size == bins
        >>> assert edges.size == bins + 1
        >>> result = get_histinfo_str(hist, edges)
        >>> print(result)

    Example1:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> data = np.array([ 0,  1,  2,  3.5,  3,  3,  4,  4])
        >>> weights = np.array([4.5, 1., 1., 1., 1., 1., 1., 1.])
        >>> range_ = (-.5, 4.5)
        >>> bins = 5
        >>> interpolation_wrap = True
        >>> hist, edges = interpolated_histogram(data, weights, range_, bins,
        >>>                                      interpolation_wrap)
        >>> assert np.abs(hist.sum() - weights.sum()) < 1E-9
        >>> assert hist.size == bins
        >>> assert edges.size == bins + 1
        >>> result = get_histinfo_str(hist, edges)
        >>> print(result)
    """
    assert bins > 0, 'must have nonzero bins'
    data = np.asarray(data)
    if weights is not None:
        weights = np.asarray(weights)
        assert np.all(weights.shape == data.shape), 'shapes disagree'
        weights = weights.ravel()
    data = data.ravel()
    # Compute bin edges like in np.histogram
    start, stop = float(range_[0]), float(range_[1])
    if start == stop:
        start -= 0.5
        stop += 0.5
    # Find bin edges
    hist_dtype = np.float64
    # Compute bin step size, add one if last bin is the same as the first
    step = (stop - start) / float((bins + interpolation_wrap))
    #edges = [start + i * step for i in range(bins + 1)]
    #centers = hist_edges_to_centers(edges)

    half_step = step / 2.0
    # Find fractional bin center index for each datapoint
    data_offset = start + half_step
    frac_index  = (data - data_offset) / step
    # Find bin center to the left of each datapoint
    left_index = np.floor(frac_index).astype(np.int32)
    # Find bin center to the right of each datapoint
    right_index = left_index + 1
    # Find the fraction of the distiance the right center is away from the datapoint
    right_alpha = (frac_index - left_index)
    left_alpha = 1.0 - right_alpha

    if _debug:
        print('bins = %r' % bins)
        print('step = %r' % step)
        print('half_step = %r' % half_step)
        print('data_offset = %r' % data_offset)
        print("-.5 MOD tau = %r" % (-.5 % TAU,))

    # Handle edge cases
    if interpolation_wrap:
        # when the stop == start (like in orientations)
        left_index  %= bins
        right_index %= bins
    else:
        left_index[left_index < 0] = 0
        right_index[right_index >= bins] = bins - 1

    # Each keypoint votes into its left and right bins
    left_vote  = left_alpha * weights
    right_vote = right_alpha * weights
    hist = np.zeros((bins,), hist_dtype)
    # TODO: can problably do this faster with cumsum
    for index, vote in zip(left_index, left_vote):
        hist[index] += vote
    for index, vote in zip(right_index, right_vote):
        hist[index] += vote

    if interpolation_wrap:
        edges = np.linspace(start, stop, bins + 1, endpoint=False)
    else:
        edges = np.linspace(start, stop, bins + 1, endpoint=True)
    if _debug:
        import vtool_ibeis as vt
        assert np.allclose(np.diff(edges), step)
        print(hist.shape)
        print(edges.shape)
        print(vt.kpts_docrepr(hist, 'hist', False))
        print(vt.kpts_docrepr(edges, 'edges', False))

    return hist, edges


def hist_edges_to_centers(edges):
    r"""
    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> edges = [-0.79, 0.00, 0.79, 1.57, 2.36, 3.14, 3.93, 4.71, 5.50, 6.28, 7.07]
        >>> centers = hist_edges_to_centers(edges)
        >>> result = str(centers)
        >>> print(result)
        [-0.395  0.395  1.18   1.965  2.75   3.535  4.32   5.105  5.89   6.675]
    """
    centers = np.array([(e1 + e2) / 2.0 for (e1, e2) in zip(edges[:-1], edges[1:])])
    return centers


def wrap_histogram(hist_, edges_, _debug=False):
    r"""
    Simulates the first and last histogram bin being being adjacent to one another
    by replicating those bins at the last and first positions respectively.

    Args:
        hist_ (ndarray):
        edges_ (ndarray):

    Returns:
        tuple: (hist_wrap, edge_wrap)

    CommandLine:
        python -m vtool_ibeis.histogram --test-wrap_histogram

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> hist_ = np.array([8., 0., 0., 34.32, 29.45, 0., 0., 6.73])
        >>> edges_ = np.array([ 0.        ,  0.78539816,  1.57079633,
        ...                    2.35619449,  3.14159265,  3.92699081,
        ...                    4.71238898,  5.49778714,  6.2831853 ])
        >>> (hist_wrap, edge_wrap) = wrap_histogram(hist_, edges_)
        >>> tup = (hist_wrap.tolist(), edge_wrap.tolist())
        >>> import ubelt as ub
        >>> result = ub.repr2(tup, nl=1, nobr=True, precision=2)
        >>> print(result)

        [6.73, 8.00, 0.00, 0.00, 34.32, 29.45, 0.00, 0.00, 6.73, 8.00],
        [-0.79, 0.00, 0.79, 1.57, 2.36, 3.14, 3.93, 4.71, 5.50, 6.28, 7.07],
    """
    # FIXME; THIS NEEDS INFORMATION ABOUT THE DISTANCE FROM THE LAST BIN
    # TO THE FIRST. IT IS OK AS LONG AS ALL STEPS ARE EQUAL, BUT IT IS NOT
    # GENERAL
    left_step, right_step = np.diff(edges_)[[0, -1]]
    hist_wrap = np.hstack((hist_[-1:], hist_, hist_[0:1]))
    edge_wrap = np.hstack((edges_[0:1] - left_step, edges_, edges_[-1:] + right_step))
    if _debug:
        import vtool_ibeis as vt
        print(vt.kpts_docrepr(hist_wrap, 'hist_wrap', False))
        print(vt.kpts_docrepr(edge_wrap, 'edge_wrap', False))
    return hist_wrap, edge_wrap


def maxima_neighbors(argmaxima, hist_, centers=None):
    neighbs = np.vstack((argmaxima - 1, argmaxima, argmaxima + 1))
    y123 = hist_[neighbs]
    x123 = neighbs if centers is None else centers[neighbs]
    return x123, y123


def maximum_parabola_point(A, B, C):
    """ Maximum x point is where the derivative is 0 """
    xv = -B / (2 * A)
    yv = C - B * B / (4 * A)
    return xv, yv


def subbin_bounds(z, radius, low, high):
    """
    Gets quantized bounds of a sub-bin/pixel point and a radius.
    Useful for cropping using subpixel points

    Illustration::
        (the bin edges are pipes)
        (the bin centers are pluses)

        Input = {'z': 1.5, 'radius':5.666, 'low':0, 'high':7}
        Output = {'z1':0, 'z2': 7, 'offst': 5.66}

        |   |   |   |   |   |   |   |   |
        |_+_|_+_|_+_|_+_|_+_|_+_|_+_|_+_|
          ^     ^                     ^
          z1    z                     z2
                ,.___.___.___.___.___.   < radius (5.333)
          .---.-,                        < z_offset1 (1.6666)
                ,_.___.___.___.___.___.  < z_offset2 (5.666)
          .---.-,                        < z_offset1 (1.6666)

    Args:
        z (float): center of a circle a 1d pixel array
        radius (float): radius of the circle
        low (int): minimum index of 1d pixel array
        high (int): maximum index of 1d pixel array

    Returns:
        tuple: (iz1, iz2, z_offst) - quantized_bounds and subbin_offset
            iz1 - low radius endpoint
            iz2 - high radius endpoint
            z_offst - subpixel offset
            #Returns: quantized_bounds=(iz1, iz2), subbin_offset

    CommandLine:
        python -m vtool_ibeis.histogram --test-subbin_bounds

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> z = 1.5
        >>> radius = 5.666
        >>> low = 0
        >>> high = 7
        >>> (iz1, iz2, z_offst) = subbin_bounds(z, radius, low, high)
        >>> result = str((iz1, iz2, z_offst))
        >>> print(result)
        (0, 7, 1.5)
    """
    #print('quan pxl: z=%r, radius=%r, low=%r, high=%r' % (z, radius, low, high))
    # Get subpixel bounds ignoring boundaries
    z1 = z - radius
    z2 = z + radius
    # Quantize and clip bounds
    iz1 = int(max(np.floor(z1), low))
    iz2 = int(min(np.ceil(z2), high))
    # Quantized min radius
    z_offst = z - iz1
    return iz1, iz2, z_offst


def show_ori_image_ondisk():
    r"""
    CommandLine:
        python -m vtool_ibeis.histogram --test-show_ori_image_ondisk --show

        python -m vtool_ibeis.histogram --test-show_ori_image_ondisk --show --patch_img_fpath patches/KP_0_PATCH.png --ori_img_fpath patches/KP_0_orientations01.png --weights_img_fpath patches/KP_0_WEIGHTS.png --grady_img_fpath patches/KP_0_ygradient.png --gradx_img_fpath patches/KP_0_xgradient.png --title cpp_show_ori_ondisk

        python -m pyhesaff._pyhesaff --test-test_rot_invar --show --rebuild-hesaff --no-rmbuild

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.histogram import *  # NOQA
        >>> import plottool_ibeis as pt
        >>> import vtool_ibeis as vt
        >>> result = show_ori_image_ondisk()
        >>> pt.show_if_requested()
    """
    #if img_fpath is not None:
    #    img_fpath = ut.get_argval('--fpath', type_=str, default=ut.grab_test_imgpath('star.png'))
    #    img_fpath = ut.get_argval('--fpath', type_=str, default=ut.grab_test_imgpath('star.png'))
    #    img = vt.imread(img_fpath)
    #    ori_img_fpath     = ut.get_argval('--fpath-ori', type_=str,
    #    default=ut.augpath(img_fpath, '_ori'))
    #    weights_img_fpath = ut.get_argval('--fpath-weight', type_=str,
    #    default=ut.augpath(img_fpath, '_mag'))
    #    vt.imwrite(ori_img_fpath, vt.patch_ori(*vt.patch_gradient(img)))
    #    vt.imwrite(weights_img_fpath, vt.patch_mag(*vt.patch_gradient(img)))
    import vtool_ibeis as vt
    print('show_ori_image_ondisk')
    def parse_img_from_arg(argstr_):
        fpath = ut.get_argval(argstr_, type_=str, default='None')
        if fpath is not None and fpath != 'None':
            img = vt.imread(fpath, grayscale=True)
            print('Reading %s with stats %s' % (fpath, ut.get_stats_str(img, axis=None)))
        else:
            print('Did not read %s' % (fpath))
            img = None
        return img
    patch = parse_img_from_arg('--patch_img_fpath')
    gori  = parse_img_from_arg('--ori_img_fpath') / 255.0 * TAU
    weights = parse_img_from_arg('--weights_img_fpath') / 255.0
    gradx = parse_img_from_arg('--gradx_img_fpath') / 255.0
    grady = parse_img_from_arg('--grady_img_fpath') / 255.0
    gauss = parse_img_from_arg('--gauss_weights_img_fpath') / 255.0
    #print(' * ori_img_fpath = %r' % (ori_img_fpath,))
    #print(' * weights_img_fpath = %r' % (weights_img_fpath,))
    #print(' * gradx_img_fpath = %r' % (gradx_img_fpath,))
    #print(' * grady_img_fpath = %r' % (grady_img_fpath,))
    #import cv2
    #cv2.imread(ori_img_fpath, cv2.IMREAD_UNCHANGED)
    show_ori_image(gori, weights, patch, gradx, grady, gauss)
    title = ut.get_argval('--title', default='')
    import plottool_ibeis as pt
    pt.set_figtitle(title)


def show_ori_image(gori, weights, patch, gradx=None, grady=None, gauss=None, fnum=None):
    """
    CommandLine:
        python -m pyhesaff._pyhesaff --test-test_rot_invar --show --nocpp
    """
    import plottool_ibeis as pt
    if fnum is None:
        fnum = pt.next_fnum()
    print('gori.max = %r' % gori.max())
    assert gori.max() <= TAU
    assert gori.min() >= 0
    bgr_ori = pt.color_orimag(gori, weights, False, encoding='bgr')
    print('bgr_ori.max = %r' % bgr_ori.max())

    bgr_ori = (255 * bgr_ori).astype(np.uint8)
    print('bgr_ori.max = %r' % bgr_ori.max())
    #bgr_ori = np.array(bgr_ori, dtype=np.uint8)
    legend = pt.make_ori_legend_img()
    #gorimag_, woff, hoff = vt.stack_images(bgr_ori, legend, vert=False, modifysize=True)
    import vtool_ibeis as vt
    gorimag_, offsets, sftup = vt.stack_images(bgr_ori, legend, vert=False,
                                               modifysize=True,
                                               return_sf=True)
    (woff, hoff) = offsets[1]
    if patch is None:
        pt.imshow(gorimag_, fnum=fnum)
    else:
        pt.imshow(gorimag_, fnum=fnum, pnum=(3, 1, 1), title='colored by orientation')
        #pt.imshow(patch, fnum=fnum, pnum=(2, 2, 1))
        #gradx, grady = np.cos(gori + TAU / 4.0), np.sin(gori + TAU / 4.0)
        if gradx is not None and grady is not None:
            if weights is not None:
                gradx *= weights
                grady *= weights
            pt.imshow(np.array(gradx * 255, dtype=np.uint8), fnum=fnum, pnum=(3, 3, 4))
            pt.imshow(np.array(grady * 255, dtype=np.uint8), fnum=fnum, pnum=(3, 3, 5))
            #pt.imshow(bgr_ori, pnum=(2, 2, 4))
            pt.draw_vector_field(gradx, grady, pnum=(3, 3, 6), invert=True)
        pt.imshow(patch, fnum=fnum, pnum=(3, 1, 3))


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.histogram
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
